import numpy as np
import openmdao.api as om

from ...utils.misc import get_rate_units
from ...utils.lgl import lgl

from scipy.interpolate import CubicSpline


class CubicSplineControlInterpComp(om.ExplicitComponent):
    """
    A component which interpolates control values in 1D using cubic spline interpolation.

    Takes training values for control variables at given _input_ nodes,
    broadcasts them to _discretization_ nodes, and then interpolates the discretization values
    to provide a control variable at a given segment tau or phase tau.

    This interpolation method is intended for use in simulation phases and should not be used to
    solve optimization problems.

    For dynamic controls, the current segment is given as a discrete input and the interpolation is
    a smooth polynomial along the given segment.

    OpenMDAO assumes sizes of variables at setup time, and we don't want to need to change the
    size of the control input nodes when we evaluate different segments. Instead, this component
    will take in the control values of all segments and internally use the appropriate one.

    Parameters
    ----------
    grid_data : GridData
        A GridData instance that details information on how the control input and discretization
        nodes are layed out.
    control_options : dict of {str: ControlOptionsDictionary}
        A mapping that maps the name of each control to a ControlOptionsDictionary of its options.
    time_units : str
        The time units pertaining to the control rates.
    standalone_mode : bool
        If True, this component runs its configuration steps during setup. This is useful for
        unittests in which the component does not exist in a larger group.
    **kwargs
        Keyword arguments passed to ExplicitComponent.
    """
    def __init__(self, grid_data, control_options=None,
                 time_units=None, standalone_mode=False, **kwargs):
        self._grid_data = grid_data
        self._control_options = {} if control_options is None else control_options
        self._time_units = time_units
        self._standalone_mode = standalone_mode

        # Cache formatted strings: { control_name : (input_name, output_name) }
        self._control_io_names = {}
        self._polynomial_control_nodes = {}
        self.num_uhat_nodes = None
        self._input_grid = []

        super().__init__(**kwargs)

    def initialize(self):
        """
        Declare component options.
        """
        self.options.declare('segment_index', types=int, desc='index of the current segment')
        self.options.declare('vec_size', types=int, default=1,
                             desc='number of points at which the control will be evaluated. This is not'
                                  'necessarily the same as the number of nodes in the GridData.')
        self.options.declare('compute_derivs', types=bool, default=True,
                             desc='Set to True if the interpolant needs to also compute the derivatives '
                             'of the outputs, otherwise False. This should be set to False for simulation mode '
                             'and true when using ExplicitShooting for optimization')

    def _configure_controls(self):
        vec_size = self.options['vec_size']
        gd = self._grid_data

        if not self._control_options:
            return

        self.num_uhat_nodes = gd.subset_num_nodes['control_input']
        for control_name, options in self._control_options.items():
            if options['control_type'] == 'full':
                shape = options['shape']
                units = options['units']
                input_name = f'controls:{control_name}'
                output_name = f'control_values:{control_name}'
                rate_name = f'control_rates:{control_name}_rate'
                rate2_name = f'control_rates:{control_name}_rate2'
                rate_units = get_rate_units(units, self._time_units)
                rate2_units = get_rate_units(units, self._time_units, deriv=2)
                uhat_shape = (self.num_uhat_nodes,) + shape
                output_shape = (vec_size,) + shape
                self.add_input(input_name, shape=uhat_shape, units=units)
                self.add_output(output_name, shape=output_shape, units=units)
                self.add_output(rate_name, shape=output_shape, units=rate_units)
                self.add_output(rate2_name, shape=output_shape, units=rate2_units)
                self._control_io_names[control_name] = (input_name, output_name, rate_name, rate2_name)

                input_ptau_grid = self._grid_data.node_ptau[self._grid_data.subset_node_indices['control_input']]
                _, self._input_grid = np.unique(input_ptau_grid, return_index=True)

            else:
                order = options['order']
                shape = options['shape']
                units = options['units']
                input_name = f'controls:{control_name}'
                output_name = f'control_values:{control_name}'
                rate_name = f'control_rates:{control_name}_rate'
                rate2_name = f'control_rates:{control_name}_rate2'
                rate_units = get_rate_units(units, self._time_units)
                rate2_units = get_rate_units(units, self._time_units, deriv=2)
                input_shape = (order + 1,) + shape
                output_shape = (vec_size,) + shape
                self.add_input(input_name, shape=input_shape, units=units)
                self.add_output(output_name, shape=output_shape, units=units)
                self.add_output(rate_name, shape=output_shape, units=rate_units)
                self.add_output(rate2_name, shape=output_shape, units=rate2_units)
                self._control_io_names[control_name] = (input_name, output_name, rate_name, rate2_name)
                self._polynomial_control_nodes[control_name], _ = lgl(order+1)

    def setup(self):
        """
        Perform the I/O creation if operating in _standalone_mode.
        """
        if self._standalone_mode:
            self.configure_io()

    def set_segment_index(self, idx, **kwargs):
        """
        Set the active segment index for control interpolation.

        Parameters
        ----------
        idx : int
            The index of the segment in which the controls are to be interpolated.
        **kwargs : dict, optional
            Keyword arguments that make this interpolant call-compatible with the BarycentricLagrangeInterpolant.
        """
        self.options['segment_index'] = idx

    def configure_io(self):
        """
        I/O creation is delayed until configure so we can determine shape and units for the controls.
        """
        vec_size = self.options['vec_size']

        self.add_input('stau', shape=(vec_size,), units=None)
        self.add_input('dstau_dt', val=1.0, units=f'1/{self._time_units}')
        self.add_input('t_duration', val=1.0, units=self._time_units)
        self.add_input('ptau', shape=(vec_size,), units=None)

        self._configure_controls()

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        """
        Compute interpolated control values and rates.

        Parameters
        ----------
        inputs : `Vector`
            `Vector` containing inputs.
        outputs : `Vector`
            `Vector` containing outputs.
        discrete_inputs : `Vector`
            `Vector` containing discrete_inputs.
        discrete_outputs : `Vector`
            `Vector` containing discrete_outputs.
        """
        vec_size = self.options['vec_size']
        ptau = inputs['ptau']
        ptau_grid = self._grid_data.node_ptau

        if self._control_options:
            input_node_idxs = self._grid_data.subset_node_indices['control_input']

            for control_name, options in self._control_options.items():
                if options['control_type'] == 'full':
                    input_name, output_name, rate_name, rate2_name = self._control_io_names[control_name]
                    shape = options['shape']

                    size = np.prod(shape)
                    out = np.zeros((vec_size, size))
                    rate = np.zeros((vec_size, size))
                    rate2 = np.zeros((vec_size, size))

                    for i in range(size):
                        spl = CubicSpline(ptau_grid[input_node_idxs][self._input_grid],
                                          inputs[input_name][self._input_grid].flatten('F')[self.num_uhat_nodes*i:
                                                                                            self.num_uhat_nodes*(i+1)])
                        out[:, i] = spl(ptau)
                        rate[:, i] = spl(ptau, nu=1) / (0.5 * inputs['t_duration'])
                        rate2[:, i] = spl(ptau, nu=2) / (0.5 * inputs['t_duration'])**2

                    outputs[output_name] = out.reshape((vec_size,) + shape, order='F')
                    outputs[rate_name] = rate.reshape((vec_size,) + shape, order='F')
                    outputs[rate2_name] = rate2.reshape((vec_size,) + shape, order='F')
                else:
                    input_name, output_name, rate_name, rate2_name = self._control_io_names[control_name]
                    order = options['order']
                    poly = np.polyfit(self._polynomial_control_nodes[control_name], inputs[input_name].ravel(), order)
                    der1 = np.polyder(poly)
                    der2 = np.polyder(der1)
                    outputs[output_name] = np.polyval(poly, ptau)
                    outputs[rate_name] = np.polyval(der1, ptau) / (0.5 * inputs['t_duration'])
                    outputs[rate2_name] = np.polyval(der2, ptau) / (0.5 * inputs['t_duration'])**2



import unittest

import numpy as np

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse

import dymos as dm
from dymos.utils.misc import om_version
from dymos.utils.testing_utils import assert_timeseries_near_equal
from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE


@use_tempdirs
class TestIntegrateControl(unittest.TestCase):

    def _test_integrate_control(self, transcription):

        #
        # Define the OpenMDAO problem
        #
        p = om.Problem(model=om.Group())

        #
        # Define a Trajectory object
        #
        traj = dm.Trajectory()

        p.model.add_subsystem('traj', subsys=traj)

        #
        # Define a Dymos Phase object with GaussLobatto Transcription
        #
        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=transcription(num_segments=10, order=3))

        traj.add_phase(name='phase0', phase=phase)

        #
        # Set the time options
        # Time has no targets in our ODE.
        # We fix the initial time so that the it is not a design variable in the optimization.
        # The duration of the phase is allowed to be optimized, but is bounded on [0.5, 10].
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(1.0, 10.0), units='s')

        #
        # Set the time options
        # Initial values of positions and velocity are all fixed.
        # The final value of position are fixed, but the final velocity is a free variable.
        # The equations of motion are not functions of position, so 'x' and 'y' have no targets.
        # The rate source points to the output in the ODE which provides the time derivative of the
        # given state.
        phase.add_state('x', fix_initial=True, fix_final=True)
        phase.add_state('y', fix_initial=True, fix_final=True)
        phase.add_state('v', fix_initial=True)

        phase.add_state('int_theta', fix_initial=False, rate_source='theta_rate', targets=['theta'])

        # Define theta as a control.
        phase.add_control(name='theta_rate', units='rad/s', shape=(1,), targets=None)

        # Minimize final time.
        phase.add_objective('time', loc='final')

        # Set the driver.
        p.driver = om.ScipyOptimizeDriver()

        # Allow OpenMDAO to automatically determine our sparsity pattern.
        # Doing so can significant speed up the execution of Dymos.
        p.driver.declare_coloring()

        # Setup the problem
        p.setup(check=True)

        # Now that the OpenMDAO problem is setup, we can set the values of the states.

        phase.set_time_val(initial=0.0, duration=5.0, units='s')

        phase.set_state_val('x', [0, 10], units='m')
        phase.set_state_val('y', [10, 5], units='m')
        phase.set_state_val('v', [0, 5], units='m/s')
        phase.set_state_val('int_theta', [0.1, 45], units='deg')
        phase.set_control_val('theta_rate', [10, 10], units='deg/s')

        # Run the driver to solve the problem
        dm.run_problem(p, simulate=True, make_plots=False)

        sol_db = 'dymos_solution.db'
        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = p.get_outputs_dir() / sol_db
            sim_db = traj.sim_prob.get_outputs_dir() / sim_db

        sol = om.CaseReader(sol_db).get_case('final')
        sim = om.CaseReader(sim_db).get_case('final')

        t_sol = sol.get_val('traj.phase0.timeseries.time')
        t_sim = sim.get_val('traj.phase0.timeseries.time')

        x_sol = sol.get_val('traj.phase0.timeseries.x')
        x_sim = sim.get_val('traj.phase0.timeseries.x')

        y_sol = sol.get_val('traj.phase0.timeseries.y')
        y_sim = sim.get_val('traj.phase0.timeseries.y')

        v_sol = sol.get_val('traj.phase0.timeseries.v')
        v_sim = sim.get_val('traj.phase0.timeseries.v')

        int_theta_sol = sol.get_val('traj.phase0.timeseries.int_theta')
        int_theta_sim = sim.get_val('traj.phase0.timeseries.int_theta')

        theta_rate_sol = sol.get_val('traj.phase0.timeseries.theta_rate')
        theta_rate_sim = sim.get_val('traj.phase0.timeseries.theta_rate')

        assert_timeseries_near_equal(t_sol, x_sol, t_sim, x_sim, rel_tolerance=4.0E-3, abs_tolerance=1.0E-2)
        assert_timeseries_near_equal(t_sol, y_sol, t_sim, y_sim, rel_tolerance=4.0E-3, abs_tolerance=1.0E-2)
        assert_timeseries_near_equal(t_sol, v_sol, t_sim, v_sim, rel_tolerance=4.0E-3, abs_tolerance=1.0E-2)
        assert_timeseries_near_equal(t_sol, int_theta_sol, t_sim, int_theta_sim, rel_tolerance=4.0E-3, abs_tolerance=1.0E-2)
        assert_timeseries_near_equal(t_sim, theta_rate_sim, t_sol, theta_rate_sol,  rel_tolerance=4.0E-3, abs_tolerance=1.0E-2)

    def _test_integrate_control_rate(self, transcription):

        #
        # Define the OpenMDAO problem
        #
        p = om.Problem(model=om.Group())

        #
        # Define a Trajectory object
        #
        traj = dm.Trajectory()

        p.model.add_subsystem('traj', subsys=traj)

        #
        # Define a Dymos Phase object with GaussLobatto Transcription
        #
        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=transcription(num_segments=10, order=3))

        traj.add_phase(name='phase0', phase=phase)

        #
        # Set the time options
        # Time has no targets in our ODE.
        # We fix the initial time so that the it is not a design variable in the optimization.
        # The duration of the phase is allowed to be optimized, but is bounded on [0.5, 10].
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(1.0, 10.0), units='s')

        #
        # Set the time options
        # Initial values of positions and velocity are all fixed.
        # The final value of position are fixed, but the final velocity is a free variable.
        # The equations of motion are not functions of position, so 'x' and 'y' have no targets.
        # The rate source points to the output in the ODE which provides the time derivative of the
        # given state.
        phase.add_state('x', fix_initial=True, fix_final=True)
        phase.add_state('y', fix_initial=True, fix_final=True)
        phase.add_state('v', fix_initial=True)

        phase.add_state('int_theta', fix_initial=False, rate_source='theta_rate', targets=['theta'])

        # Define theta as a control.
        phase.add_control(name='theta', units='rad', shape=(1,), targets=None)

        # Force the initial value of the theta polynomial control to equal the initial value of the theta state.
        traj.add_linkage_constraint(phase_a='phase0', phase_b='phase0',
                                    var_a='theta', var_b='int_theta',
                                    loc_a='initial', loc_b='initial')

        # Minimize final time.
        phase.add_objective('time', loc='final')

        # Set the driver.
        p.driver = om.ScipyOptimizeDriver()

        # Allow OpenMDAO to automatically determine our sparsity pattern.
        # Doing so can significant speed up the execution of Dymos.
        p.driver.declare_coloring()

        # Setup the problem
        p.setup(check=True)

        # Now that the OpenMDAO problem is setup, we can set the values of the states.

        phase.set_time_val(initial=0.0, duration=5.0, units='s')

        phase.set_state_val('x', [0, 10], units='m')
        phase.set_state_val('y', [10, 5], units='m')
        phase.set_state_val('v', [0, 5], units='m/s')
        phase.set_state_val('int_theta', [0.1, 45], units='deg')
        phase.set_control_val('theta', [0, 100], units='deg')

        # Run the driver to solve the problem
        dm.run_problem(p, simulate=True, make_plots=False)

        sol_db = 'dymos_solution.db'
        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = p.get_outputs_dir() / sol_db
            sim_db = traj.sim_prob.get_outputs_dir() / sim_db

        sol = om.CaseReader(sol_db).get_case('final')
        sim = om.CaseReader(sim_db).get_case('final')

        t_sol = sol.get_val('traj.phase0.timeseries.time')
        t_sim = sim.get_val('traj.phase0.timeseries.time')

        x_sol = sol.get_val('traj.phase0.timeseries.x')
        x_sim = sim.get_val('traj.phase0.timeseries.x')

        y_sol = sol.get_val('traj.phase0.timeseries.y')
        y_sim = sim.get_val('traj.phase0.timeseries.y')

        v_sol = sol.get_val('traj.phase0.timeseries.v')
        v_sim = sim.get_val('traj.phase0.timeseries.v')

        int_theta_sol = sol.get_val('traj.phase0.timeseries.int_theta')
        int_theta_sim = sim.get_val('traj.phase0.timeseries.int_theta')

        theta_sol = sol.get_val('traj.phase0.timeseries.theta')
        theta_sim = sim.get_val('traj.phase0.timeseries.theta')

        assert_timeseries_near_equal(t_sol, x_sol, t_sim, x_sim, rel_tolerance=4.0E-3, abs_tolerance=1.0E-2)
        assert_timeseries_near_equal(t_sol, y_sol, t_sim, y_sim, rel_tolerance=4.0E-3, abs_tolerance=1.0E-2)
        assert_timeseries_near_equal(t_sol, v_sol, t_sim, v_sim, rel_tolerance=4.0E-3, abs_tolerance=1.0E-2)
        assert_timeseries_near_equal(t_sol, int_theta_sol, t_sim, int_theta_sim, rel_tolerance=4.0E-3, abs_tolerance=1.0E-2)

        assert_timeseries_near_equal(t_sol, int_theta_sol, t_sol, theta_sol, rel_tolerance=4.0E-3, abs_tolerance=1.0E-2)
        assert_timeseries_near_equal(t_sim, int_theta_sim, t_sim, theta_sim, rel_tolerance=4.0E-3, abs_tolerance=1.0E-2)

    @require_pyoptsparse(optimizer='SLSQP')
    def _test_integrate_control_rate2(self, transcription):
        #
        # Define the OpenMDAO problem
        #
        p = om.Problem(model=om.Group())

        #
        # Define a Trajectory object
        #
        traj = dm.Trajectory()

        p.model.add_subsystem('traj', subsys=traj)

        #
        # Define a Dymos Phase object with GaussLobatto Transcription
        #
        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=transcription(num_segments=15, order=3))

        traj.add_phase(name='phase0', phase=phase)

        #
        # Set the time options
        # Time has no targets in our ODE.
        # We fix the initial time so that the it is not a design variable in the optimization.
        # The duration of the phase is allowed to be optimized, but is bounded on [0.5, 10].
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(1.0, 10.0), units='s')

        #
        # Set the time options
        # Initial values of positions and velocity are all fixed.
        # The final value of position are fixed, but the final velocity is a free variable.
        # The equations of motion are not functions of position, so 'x' and 'y' have no targets.
        # The rate source points to the output in the ODE which provides the time derivative of the
        # given state.
        phase.add_state('x', fix_initial=True, fix_final=True)
        phase.add_state('y', fix_initial=True, fix_final=True)
        phase.add_state('v', fix_initial=True)

        phase.add_state('int_theta_dot', fix_initial=False, rate_source='theta_rate2')
        phase.add_state('int_theta', fix_initial=False, rate_source='int_theta_dot', targets=['theta'])

        # Define theta as a control.
        phase.add_control(name='theta', units='rad', shape=(1,), targets=None)

        # Force the initial value of the theta polynomial control to equal the initial value of the theta state.
        traj.add_linkage_constraint(phase_a='phase0', phase_b='phase0',
                                    var_a='theta', var_b='int_theta',
                                    loc_a='initial', loc_b='initial')

        traj.add_linkage_constraint(phase_a='phase0', phase_b='phase0',
                                    var_a='int_theta_dot', var_b='theta_rate',
                                    loc_a='initial', loc_b='initial',
                                    units='rad/s')

        # Note we have to add theta_rate to the timeseries outputs here because
        # linkages get boundary values from that output.
        phase.add_timeseries_output('theta_rate')

        # Minimize final time.
        phase.add_objective('time', loc='final')

        # Set the driver.
        p.driver = om.pyOptSparseDriver(optimizer='SLSQP')
        p.driver.opt_settings['ACC'] = 1e-9

        # Allow OpenMDAO to automatically determine our sparsity pattern.
        # Doing so can significant speed up the execution of Dymos.
        p.driver.declare_coloring()

        # Setup the problem
        p.setup(check=True)

        # Now that the OpenMDAO problem is setup, we can set the values of the states.
        phase.set_time_val(initial=0.0, duration=5.0, units='s')

        phase.set_state_val('x', [0, 10], units='m')
        phase.set_state_val('y', [10, 5], units='m')
        phase.set_state_val('v', [0, 5], units='m/s')
        phase.set_state_val('int_theta', [0.1, 45], units='deg')
        phase.set_state_val('int_theta_dot', [0.0, 0.0], units='deg/s')
        phase.set_control_val('theta', [0, 100], units='deg')

        # Run the driver to solve the problem
        dm.run_problem(p, simulate=True, make_plots=False, simulate_kwargs={'atol': 1.0E-9, 'rtol': 1.0E-9,
                                                                            'times_per_seg': 10})

        sol_db = 'dymos_solution.db'
        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = p.get_outputs_dir() / sol_db
            sim_db = traj.sim_prob.get_outputs_dir() / sim_db

        sol = om.CaseReader(sol_db).get_case('final')
        sim = om.CaseReader(sim_db).get_case('final')

        t_sol = sol.get_val('traj.phase0.timeseries.time')
        t_sim = sim.get_val('traj.phase0.timeseries.time')

        x_sol = sol.get_val('traj.phase0.timeseries.x')
        x_sim = sim.get_val('traj.phase0.timeseries.x')

        y_sol = sol.get_val('traj.phase0.timeseries.y')
        y_sim = sim.get_val('traj.phase0.timeseries.y')

        v_sol = sol.get_val('traj.phase0.timeseries.v')
        v_sim = sim.get_val('traj.phase0.timeseries.v')

        int_theta_sol = sol.get_val('traj.phase0.timeseries.int_theta')
        int_theta_sim = sim.get_val('traj.phase0.timeseries.int_theta')

        theta_sol = sol.get_val('traj.phase0.timeseries.theta')
        theta_sim = sim.get_val('traj.phase0.timeseries.theta')

        assert_timeseries_near_equal(t_sol, x_sol, t_sim, x_sim, rel_tolerance=1.0E-2, abs_tolerance=1.0E-1)
        assert_timeseries_near_equal(t_sol, y_sol, t_sim, y_sim, rel_tolerance=1.0E-2, abs_tolerance=1.0E-1)
        assert_timeseries_near_equal(t_sol, v_sol, t_sim, v_sim, rel_tolerance=1.0E-2, abs_tolerance=1.0E-1)
        assert_timeseries_near_equal(t_sol, int_theta_sol, t_sim, int_theta_sim, rel_tolerance=1.0E-2,
                                     abs_tolerance=1.0E-1)

        assert_timeseries_near_equal(t_sol, int_theta_sol, t_sol, theta_sol, rel_tolerance=1.0E-2, abs_tolerance=1.0E-1)
        assert_timeseries_near_equal(t_sim, int_theta_sim, t_sim, theta_sim, rel_tolerance=1.0E-2, abs_tolerance=1.0E-1)

    def test_integrate_control_gl(self):
        self._test_integrate_control(dm.GaussLobatto)

    def test_integrate_control_radau(self):
        self._test_integrate_control(dm.Radau)

    def test_integrate_control_rate_gl(self):
        self._test_integrate_control_rate(dm.GaussLobatto)

    def test_integrate_control_rate_radau(self):
        self._test_integrate_control_rate(dm.Radau)

    def test_integrate_control_rate2_gl(self):
        self._test_integrate_control_rate2(dm.GaussLobatto)

    def test_integrate_control_rate2_radau(self):
        self._test_integrate_control_rate2(dm.Radau)


@use_tempdirs
class TestIntegratePolynomialControl(unittest.TestCase):

    def _test_integrate_polynomial_control(self, transcription):
        #
        # Define the OpenMDAO problem
        #
        p = om.Problem(model=om.Group())

        #
        # Define a Trajectory object
        #
        traj = dm.Trajectory()

        p.model.add_subsystem('traj', subsys=traj)

        #
        # Define a Dymos Phase object with GaussLobatto Transcription
        #
        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=transcription(num_segments=20, order=3))

        traj.add_phase(name='phase0', phase=phase)

        #
        # Set the time options
        # Time has no targets in our ODE.
        # We fix the initial time so that the it is not a design variable in the optimization.
        # The duration of the phase is allowed to be optimized, but is bounded on [0.5, 10].
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(1.0, 10.0), units='s')

        #
        # Set the time options
        # Initial values of positions and velocity are all fixed.
        # The final value of position are fixed, but the final velocity is a free variable.
        # The equations of motion are not functions of position, so 'x' and 'y' have no targets.
        # The rate source points to the output in the ODE which provides the time derivative of the
        # given state.
        phase.add_state('x', fix_initial=True, fix_final=True)
        phase.add_state('y', fix_initial=True, fix_final=True)
        phase.add_state('v', fix_initial=True)

        phase.add_state('int_theta', fix_initial=False, rate_source='theta_rate',
                        targets=['theta'])

        # Define theta as a control.
        phase.add_control(name='theta_rate', order=11, units='rad/s', shape=(1,), targets=None, control_type='polynomial')

        # Minimize final time.
        phase.add_objective('time', loc='final')

        # Set the driver.
        p.driver = om.ScipyOptimizeDriver()

        # Allow OpenMDAO to automatically determine our sparsity pattern.
        # Doing so can significant speed up the execution of Dymos.
        p.driver.declare_coloring()

        # Setup the problem
        p.setup(check=True)

        # Now that the OpenMDAO problem is setup, we can set the values of the states.

        phase.set_time_val(initial=0.0, duration=5.0, units='s')

        phase.set_state_val('x', [0, 10], units='m')
        phase.set_state_val('y', [10, 5], units='m')
        phase.set_state_val('v', [0, 5], units='m/s')
        phase.set_state_val('int_theta', [0.1, 45], units='deg')
        phase.set_control_val('theta_rate', 10.0, units='deg/s')

        # Run the driver to solve the problem
        dm.run_problem(p, simulate=True, make_plots=True)

        sol_db = 'dymos_solution.db'
        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = p.get_outputs_dir() / sol_db
            sim_db = traj.sim_prob.get_outputs_dir() / sim_db

        sol = om.CaseReader(sol_db).get_case('final')
        sim = om.CaseReader(sim_db).get_case('final')

        t_sol = sol.get_val('traj.phase0.timeseries.time')
        t_sim = sim.get_val('traj.phase0.timeseries.time')

        x_sol = sol.get_val('traj.phase0.timeseries.x')
        x_sim = sim.get_val('traj.phase0.timeseries.x')

        y_sol = sol.get_val('traj.phase0.timeseries.y')
        y_sim = sim.get_val('traj.phase0.timeseries.y')

        v_sol = sol.get_val('traj.phase0.timeseries.v')
        v_sim = sim.get_val('traj.phase0.timeseries.v')

        int_theta_sol = sol.get_val('traj.phase0.timeseries.int_theta')
        int_theta_sim = sim.get_val('traj.phase0.timeseries.int_theta')

        theta_rate_sol = sol.get_val('traj.phase0.timeseries.theta_rate')
        theta_rate_sim = sim.get_val('traj.phase0.timeseries.theta_rate')

        assert_timeseries_near_equal(t_sol, x_sol, t_sim, x_sim, rel_tolerance=4.0E-3, abs_tolerance=1.0E-2)
        assert_timeseries_near_equal(t_sol, y_sol, t_sim, y_sim, rel_tolerance=4.0E-3, abs_tolerance=1.0E-2)
        assert_timeseries_near_equal(t_sol, v_sol, t_sim, v_sim, rel_tolerance=4.0E-3, abs_tolerance=1.0E-2)
        assert_timeseries_near_equal(t_sol, int_theta_sol, t_sim, int_theta_sim, rel_tolerance=4.0E-3,
                                     abs_tolerance=1.0E-2)
        assert_timeseries_near_equal(t_sim, theta_rate_sim, t_sol, theta_rate_sol, rel_tolerance=4.0E-3,
                                     abs_tolerance=1.0E-2)

    def _test_integrate_polynomial_control_rate(self, transcription):
        #
        # Define the OpenMDAO problem
        #
        p = om.Problem(model=om.Group())

        #
        # Define a Trajectory object
        #
        traj = dm.Trajectory()

        p.model.add_subsystem('traj', subsys=traj)

        #
        # Define a Dymos Phase object with GaussLobatto Transcription
        #
        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=transcription(num_segments=10, order=5))

        traj.add_phase(name='phase0', phase=phase)

        #
        # Set the time options
        # Time has no targets in our ODE.
        # We fix the initial time so that the it is not a design variable in the optimization.
        # The duration of the phase is allowed to be optimized, but is bounded on [0.5, 10].
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(1.0, 10.0), units='s')

        #
        # Set the time options
        # Initial values of positions and velocity are all fixed.
        # The final value of position are fixed, but the final velocity is a free variable.
        # The equations of motion are not functions of position, so 'x' and 'y' have no targets.
        # The rate source points to the output in the ODE which provides the time derivative of the
        # given state.
        phase.add_state('x', fix_initial=True, fix_final=True)
        phase.add_state('y', fix_initial=True, fix_final=True)
        phase.add_state('v', fix_initial=True)

        phase.add_state('int_theta', fix_initial=False, rate_source='theta_rate',
                        targets=['theta'])

        # Define theta as a control.
        phase.add_control(name='theta', order=11, units='rad', shape=(1,), targets=None, control_type='polynomial')

        # Force the initial value of the theta polynomial control to equal the initial value of the theta state.
        traj.add_linkage_constraint(phase_a='phase0', phase_b='phase0',
                                    var_a='theta', var_b='int_theta',
                                    loc_a='initial', loc_b='initial')

        # Minimize final time.
        phase.add_objective('time', loc='final')

        # Set the driver.
        p.driver = om.ScipyOptimizeDriver()

        # Allow OpenMDAO to automatically determine our sparsity pattern.
        # Doing so can significant speed up the execution of Dymos.
        p.driver.declare_coloring()

        # Setup the problem
        p.setup(check=True)

        # Now that the OpenMDAO problem is setup, we can set the values of the states.

        phase.set_time_val(initial=0.0, duration=5.0, units='s')

        phase.set_state_val('x', [0, 10], units='m')
        phase.set_state_val('y', [10, 5], units='m')
        phase.set_state_val('v', [0, 5], units='m/s')
        phase.set_state_val('int_theta', [0.1, 45], units='deg')
        phase.set_control_val('theta', 45.0, units='deg')

        # Run the driver to solve the problem
        dm.run_problem(p, simulate=True, make_plots=False,
                       simulate_kwargs={'times_per_seg': 10})

        sol_db = 'dymos_solution.db'
        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = p.get_outputs_dir() / sol_db
            sim_db = traj.sim_prob.get_outputs_dir() / sim_db

        sol = om.CaseReader(sol_db).get_case('final')
        sim = om.CaseReader(sim_db).get_case('final')

        t_sol = sol.get_val('traj.phase0.timeseries.time')
        t_sim = sim.get_val('traj.phase0.timeseries.time')

        x_sol = sol.get_val('traj.phase0.timeseries.x')
        x_sim = sim.get_val('traj.phase0.timeseries.x')

        y_sol = sol.get_val('traj.phase0.timeseries.y')
        y_sim = sim.get_val('traj.phase0.timeseries.y')

        v_sol = sol.get_val('traj.phase0.timeseries.v')
        v_sim = sim.get_val('traj.phase0.timeseries.v')

        int_theta_sol = sol.get_val('traj.phase0.timeseries.int_theta')
        int_theta_sim = sim.get_val('traj.phase0.timeseries.int_theta')

        theta_sol = sol.get_val('traj.phase0.timeseries.theta')
        # theta_sim = sim.get_val('traj.phase0.timeseries.theta')

        assert_timeseries_near_equal(t_sol, x_sol, t_sim, x_sim, rel_tolerance=4.0E-3, abs_tolerance=1.0E-2)
        assert_timeseries_near_equal(t_sol, y_sol, t_sim, y_sim, rel_tolerance=4.0E-3, abs_tolerance=1.0E-2)
        assert_timeseries_near_equal(t_sol, v_sol, t_sim, v_sim, rel_tolerance=4.0E-3, abs_tolerance=1.0E-2)
        assert_timeseries_near_equal(t_sol, int_theta_sol, t_sim, int_theta_sim,
                                     rel_tolerance=1.0E-2, abs_tolerance=1.5E-2)

        assert_timeseries_near_equal(t_sol, int_theta_sol, t_sol, theta_sol, rel_tolerance=4.0E-3, abs_tolerance=1.0E-2)

    @require_pyoptsparse(optimizer='SLSQP')
    def _test_integrate_polynomial_control_rate2(self, transcription):
        #
        # Define the OpenMDAO problem
        #
        p = om.Problem(model=om.Group())

        #
        # Define a Trajectory object
        #
        traj = dm.Trajectory()

        p.model.add_subsystem('traj', subsys=traj)

        #
        # Define a Dymos Phase object with GaussLobatto Transcription
        #
        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=transcription(num_segments=20, order=3))

        traj.add_phase(name='phase0', phase=phase)

        #
        # Set the time options
        # Time has no targets in our ODE.
        # We fix the initial time so that the it is not a design variable in the optimization.
        # The duration of the phase is allowed to be optimized, but is bounded on [0.5, 10].
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(1.0, 10.0), units='s')

        #
        # Set the time options
        # Initial values of positions and velocity are all fixed.
        # The final value of position are fixed, but the final velocity is a free variable.
        # The equations of motion are not functions of position, so 'x' and 'y' have no targets.
        # The rate source points to the output in the ODE which provides the time derivative of the
        # given state.
        phase.add_state('x', fix_initial=True, fix_final=True)
        phase.add_state('y', fix_initial=True, fix_final=True)
        phase.add_state('v', fix_initial=True)

        phase.add_state('int_theta_dot', fix_initial=False, rate_source='theta_rate2')
        phase.add_state('int_theta', fix_initial=False, rate_source='int_theta_dot',
                        targets=['theta'])

        # Define theta as a control.
        phase.add_control(name='theta', order=11, units='rad', shape=(1,), targets=None, control_type='polynomial')

        # Force the initial value of the theta polynomial control to equal the initial value of the theta state.
        traj.add_linkage_constraint(phase_a='phase0', phase_b='phase0',
                                    var_a='theta', var_b='int_theta',
                                    loc_a='initial', loc_b='initial')

        traj.add_linkage_constraint(phase_a='phase0', phase_b='phase0',
                                    var_a='int_theta_dot', var_b='theta_rate',
                                    loc_a='initial', loc_b='initial',
                                    units='rad/s')

        # Note for this test we have to add theta rate to the timeseries outputs since it's
        # not there by default.
        phase.add_timeseries_output('theta_rate')

        # Minimize final time.
        phase.add_objective('time', loc='final')

        # Set the driver.
        p.driver = om.pyOptSparseDriver(optimizer='SLSQP')

        # Allow OpenMDAO to automatically determine our sparsity pattern.
        # Doing so can significant speed up the execution of Dymos.
        p.driver.declare_coloring()

        # Setup the problem
        p.setup(check=True)

        # Now that the OpenMDAO problem is setup, we can set the values of the states.

        phase.set_time_val(initial=0.0, duration=5.0, units='s')

        phase.set_state_val('x', [0, 10], units='m')
        phase.set_state_val('y', [10, 5], units='m')
        phase.set_state_val('v', [0, 5], units='m/s')
        phase.set_state_val('int_theta', [0.1, 45], units='deg')
        phase.set_state_val('int_theta_dot', [0.0, 0.0], units='deg/s')
        phase.set_control_val('theta', 45.0, units='deg')

        # Run the driver to solve the problem
        dm.run_problem(p, simulate=True, make_plots=True)

        sol_db = 'dymos_solution.db'
        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = p.get_outputs_dir() / sol_db
            sim_db = traj.sim_prob.get_outputs_dir() / sim_db

        sol = om.CaseReader(sol_db).get_case('final')
        sim = om.CaseReader(sim_db).get_case('final')

        t_sol = sol.get_val('traj.phase0.timeseries.time')
        t_sim = sim.get_val('traj.phase0.timeseries.time')

        x_sol = sol.get_val('traj.phase0.timeseries.x')
        x_sim = sim.get_val('traj.phase0.timeseries.x')

        y_sol = sol.get_val('traj.phase0.timeseries.y')
        y_sim = sim.get_val('traj.phase0.timeseries.y')

        v_sol = sol.get_val('traj.phase0.timeseries.v')
        v_sim = sim.get_val('traj.phase0.timeseries.v')

        int_theta_sol = sol.get_val('traj.phase0.timeseries.int_theta')
        int_theta_sim = sim.get_val('traj.phase0.timeseries.int_theta')

        assert_timeseries_near_equal(t_sol, x_sol, t_sim, x_sim, rel_tolerance=4.0E-3, abs_tolerance=1.0E-2)

        assert_timeseries_near_equal(t_sim, y_sim, t_sol, y_sol, rel_tolerance=8.0E-3, abs_tolerance=1.0E-2)
        assert_timeseries_near_equal(t_sim, v_sim, t_sol, v_sol, rel_tolerance=8.0E-3, abs_tolerance=1.0E-2)
        assert_timeseries_near_equal(t_sim, int_theta_sim, t_sol, int_theta_sol, rel_tolerance=8.0E-3, abs_tolerance=1.0E-2)
        assert_timeseries_near_equal(t_sol, int_theta_sol, t_sim, int_theta_sim, rel_tolerance=4.0E-3, abs_tolerance=1.0E-2)

    def test_integrate_polynomial_control_gl(self):
        self._test_integrate_polynomial_control(dm.GaussLobatto)

    def test_integrate_polynomial_control_radau(self):
        self._test_integrate_polynomial_control(dm.Radau)

    def test_integrate_polynomial_control_rate_gl(self):
        self._test_integrate_polynomial_control_rate(dm.GaussLobatto)

    def test_integrate_polynomial_control_rate_radau(self):
        self._test_integrate_polynomial_control_rate(dm.Radau)

    def test_integrate_polynomial_control_rate2_gl(self):
        self._test_integrate_polynomial_control_rate2(dm.GaussLobatto)

    def test_integrate_polynomial_control_rate2_radau(self):
        self._test_integrate_polynomial_control_rate2(dm.Radau)


@use_tempdirs
class TestIntegrateTimeParamAndState(unittest.TestCase):

    def _test_transcription(self, transcription=dm.GaussLobatto, time_name='time'):
        #
        # Define the OpenMDAO problem
        #
        p = om.Problem(model=om.Group())

        #
        # Define a Trajectory object
        #
        traj = dm.Trajectory()

        p.model.add_subsystem('traj', subsys=traj)

        #
        # Define a Dymos Phase object with GaussLobatto Transcription
        #
        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=transcription(num_segments=10, order=3,
                                                     solve_segments='forward'))

        traj.add_phase(name='phase0', phase=phase)

        #
        # Set the time options
        # Time has no targets in our ODE.
        # We fix the initial time so that the it is not a design variable in the optimization.
        # The duration of the phase is allowed to be optimized, but is bounded on [0.5, 10].
        #
        phase.set_time_options(fix_initial=True, fix_duration=True, units='s', name=time_name)

        #
        # Set the time options
        # Initial values of positions and velocity are all fixed.
        # The final value of position are fixed, but the final velocity is a free variable.
        # The equations of motion are not functions of position, so 'x' and 'y' have no targets.
        # The rate source points to the output in the ODE which provides the time derivative of the
        # given state.
        phase.add_state('x', fix_initial=True)
        phase.add_state('y', fix_initial=True)
        phase.add_state('v', fix_initial=True)

        phase.add_state('int_one', fix_initial=True, rate_source='one')
        phase.add_state('int_time', fix_initial=True, rate_source=time_name)
        phase.add_state('int_time_phase', fix_initial=True, rate_source=f'{time_name}_phase')
        phase.add_state('int_int_one', fix_initial=True, rate_source='int_one')

        # Define theta as a control.
        phase.add_control(name='theta', units='rad', lower=0, upper=np.pi)

        # With no targets we must explicitly assign units and shape to this parameter.
        # Its only purpose is to be integrated as the rate source for a state.
        phase.add_parameter(name='one', opt=False, units=None, shape=(1,))

        # Minimize final time.
        phase.add_objective(time_name, loc='final')

        # Set the driver.
        p.driver = om.ScipyOptimizeDriver()

        # Allow OpenMDAO to automatically determine our sparsity pattern.
        # Doing so can significant speed up the execution of Dymos.
        p.driver.declare_coloring()

        # Setup the problem
        p.setup(check=True)

        # Now that the OpenMDAO problem is setup, we can set the values of the states.

        phase.set_time_val(initial=0.0, duration=5.0, units='s')

        phase.set_state_val('x', [0, 10], units='m')
        phase.set_state_val('y', [10, 5], units='m')
        phase.set_state_val('v', [0, 5], units='m/s')
        phase.set_state_val('int_one', [0, 10], units='s')
        phase.set_state_val('int_int_one', [0, 10], units='s**2')
        phase.set_state_val('int_time', [0, 10], units='s**2')
        phase.set_state_val('int_time_phase', [0, 10], units='s**2')
        phase.set_control_val('theta', [0.1, 45], units='deg')
        phase.set_parameter_val('one', 1.0)

        # Run the driver to solve the problem
        dm.run_problem(p, simulate=True)

        time_sol = p.get_val(f'traj.phase0.timeseries.{time_name}')
        time_phase_sol = p.get_val(f'traj.phase0.timeseries.{time_name}_phase')
        int_one_sol = p.get_val('traj.phase0.timeseries.int_one')
        int_time_sol = p.get_val('traj.phase0.timeseries.int_time')
        int_time_phase_sol = p.get_val('traj.phase0.timeseries.int_time_phase')
        int_int_one_sol = p.get_val('traj.phase0.timeseries.int_int_one')

        time_sim = p.get_val(f'traj.phase0.timeseries.{time_name}')
        time_phase_sim = p.get_val(f'traj.phase0.timeseries.{time_name}_phase')
        int_one_sim = p.get_val('traj.phase0.timeseries.int_one')
        int_time_sim = p.get_val('traj.phase0.timeseries.int_time')
        int_time_phase_sim = p.get_val('traj.phase0.timeseries.int_time_phase')
        int_int_one_sim = p.get_val('traj.phase0.timeseries.int_int_one')

        # Integral of one should match time and time_phase in this case.
        assert_near_equal(int_one_sol, time_sol, tolerance=1.0E-12)
        assert_near_equal(int_one_sol, time_phase_sol, tolerance=1.0E-12)

        assert_near_equal(int_one_sim, time_sim, tolerance=1.0E-12)
        assert_near_equal(int_one_sim, time_phase_sim, tolerance=1.0E-12)

        # Integral of time and time_phase should be t**2/2
        assert_near_equal(time_sol, time_phase_sol, tolerance=1.0E-12)
        assert_near_equal(int_time_sol, time_sol**2/2, tolerance=1.0E-12)
        assert_near_equal(int_time_phase_sol, time_phase_sol**2/2, tolerance=1.0E-12)

        assert_near_equal(time_sim, time_phase_sim, tolerance=1.0E-12)
        assert_near_equal(int_time_sim, time_sim**2/2, tolerance=1.0E-12)
        assert_near_equal(int_time_phase_sim, time_phase_sim**2/2, tolerance=1.0E-12)

        # Double integral of one should be the same as the integral of time
        assert_near_equal(int_int_one_sol, int_time_sol, tolerance=1.0E-12)
        assert_near_equal(int_int_one_sim, int_time_sim, tolerance=1.0E-12)

        assert_timeseries_near_equal(time_sol, int_int_one_sol, time_sim, int_int_one_sim, rel_tolerance=1.0E-12)

    def test_integrated_times_params_and_states(self):
        for tx in (dm.GaussLobatto, dm.Radau):
            tx_name = 'GaussLobatto' if tx is dm.GaussLobatto else 'Radau'
            for time_name in ('time', 'elapsed_time'):
                with self.subTest(msg=f'{tx_name}: time_name=\'{time_name}\''):
                    self._test_transcription(transcription=tx, time_name=time_name)


@use_tempdirs
class TestInvalidStateRateSource(unittest.TestCase):

    def test_brach_invalid_state_rate_source(self):

        class _BrachODE(om.ExplicitComponent):

            def initialize(self):
                self.options.declare('num_nodes', types=int)

            def setup(self):
                nn = self.options['num_nodes']

                # Inputs
                self.add_input('v', val=np.zeros(nn), desc='velocity', units='m/s')
                self.add_input('g', val=9.80665 * np.ones(nn), desc='grav. acceleration', units='m/s/s')
                self.add_input('theta', val=np.ones(nn), desc='angle of wire', units='rad')

                self.add_output('xdot', val=np.zeros(nn), desc='velocity component in x', units='m/s')

                self.add_output('ydot', val=np.zeros(nn), desc='velocity component in y', units='m/s')

                self.add_output('vdot', val=np.zeros(nn), desc='acceleration magnitude', units='m/s**2')

                self.add_output('check', val=np.zeros(nn), desc='check solution: v/sin(theta) = constant',
                                units='m/s')

                self.declare_coloring(wrt='*', method='cs')

            def compute(self, inputs, outputs):
                theta = inputs['theta']
                cos_theta = np.cos(theta)
                sin_theta = np.sin(theta)
                g = inputs['g']
                v = inputs['v']

                outputs['vdot'] = g * cos_theta
                outputs['xdot'] = v * sin_theta
                outputs['ydot'] = -v * cos_theta
                outputs['check'] = v / sin_theta

        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()

        t = dm.Radau(num_segments=10, order=3)

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=_BrachODE, transcription=t)
        p.model.add_subsystem('traj0', traj)
        traj.add_phase('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=False, rate_source='xdot')
        phase.add_state('y', fix_initial=True, fix_final=False, rate_source='ydot')

        # Intentionally incorrect rate source to trigger an error during configure.
        phase.add_state('v', fix_initial=True, fix_final=False, rate_source='vel_dot')

        phase.add_control('theta',
                          continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2')

        phase.add_boundary_constraint('x', loc='final', equals=10)
        phase.add_boundary_constraint('y', loc='final', equals=5)

        # Minimize time at the end of the phase
        phase.add_objective('time_phase', loc='final', scaler=10)

        with self.assertRaises(RuntimeError) as ctx:
            p.setup()

        expected = 'Error during configure_states_introspection in phase traj0.phases.phase0.'
        self.assertEqual(str(ctx.exception), expected)



import numpy as np
import openmdao.api as om
from openmdao.utils.units import unit_conversion
from scipy import sparse as sp

from ...transcriptions.grid_data import GridData
from ..._options import options as dymos_options
from ...utils.lagrange import lagrange_matrices


class TimeseriesOutputComp(om.ExplicitComponent):
    """
    Class definition of the TimeseriesOutputComp.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._no_check_partials = not dymos_options['include_check_partials']

        # _vars keeps track of the name of each output and maps to its metadata;
        # a tuple of (input_name, name, shape, rate)
        self._vars = {}

        # _sources is used internally to map the source of a connection to the timeseries to
        # the corresponding input variable.  This is used to ensure that we don't need to connect
        # the same source to this timeseries multiple times.
        self._sources = {}

        # the number of nodes from the source (this is generally all nodes)
        self.input_num_nodes = 0

        # the number of nodes in the output
        self.output_num_nodes = 0

        # Used to track conversion factors for instances when one output that relies on an input
        # from another variable has potentially different units
        self._units = {}
        self._conversion_factors = {}

        # Flag to set if no multiplication by the interpolation matrix is necessary
        self._no_interp = False

    def initialize(self):
        """
        Declare component options.
        """
        self.options.declare('input_grid_data',
                             types=GridData,
                             desc='Container object for grid on which inputs are provided.')

        self.options.declare('output_grid_data',
                             types=GridData,
                             allow_none=True,
                             default=None,
                             desc='Container object for grid on which outputs are interpolated.')

        self.options.declare('output_subset',
                             types=str,
                             default='all',
                             desc='Name of the node subset at which outputs are desired.')

        self.options.declare('time_units', default=None, allow_none=True, types=str,
                             desc='Units of time')

    def setup(self):
        """
        Define the independent variables as output variables.
        """

        # igd = self.options['input_grid_data']
        # ogd = self.options['output_grid_data']
        # output_subset = self.options['output_subset']

        # if ogd is None:
        #     ogd = igd

        # if ogd == igd and output_subset == 'all':
        #     self._no_interp = True

        # self.input_num_nodes = igd.num_nodes
        # self.output_num_nodes = ogd.subset_num_nodes[output_subset]

        # # Build the interpolation matrix which maps from the input grid to the output grid.
        # # Rather than a single phase-wide interpolating polynomial, map each segment.
        # # To do this, find the nodes in the output grid which fall in each segment of the input
        # # grid.  Then build a Lagrange interpolating polynomial for that segment
        # L_blocks = []
        # D_blocks = []
        # output_nodes_ptau = ogd.node_ptau[ogd.subset_node_indices[output_subset]]

        # for iseg in range(igd.num_segments):
        #     i1, i2 = igd.segment_indices[iseg]
        #     iptau_segi = igd.node_ptau[i1:i2]
        #     istau_segi = igd.node_stau[i1:i2]

        #     # The indices of the output grid that fall within this segment of the input grid
        #     if ogd is igd and output_subset == 'all':
        #         optau_segi = iptau_segi
        #     else:
        #         ptau_hi = igd.segment_ends[iseg+1]
        #         if iseg < igd.num_segments - 1:
        #             optau_segi = output_nodes_ptau[output_nodes_ptau <= ptau_hi]
        #         else:
        #             optau_segi = output_nodes_ptau

        #         # Remove the captured nodes so we don't accidentally include them again
        #         output_nodes_ptau = output_nodes_ptau[len(optau_segi):]

        #     # # Now get the output nodes which fall in iseg in iseg's segment tau space.
        #     ostau_segi = 2.0 * (optau_segi - iptau_segi[0]) / (iptau_segi[-1] - iptau_segi[0]) - 1

        #     # Create the interpolation matrix and add it to the blocks
        #     L, D = lagrange_matrices(istau_segi, ostau_segi)
        #     L_blocks.append(L)
        #     D_blocks.append(D)

        # self.interpolation_matrix = sp.block_diag(L_blocks, format='csr')
        # self.differentiation_matrix = sp.block_diag(D_blocks, format='csr')

    def _add_output_configure(self, name, units, shape, desc='', src=None, rate=False):
        """
        Add a single timeseries output.

        Can be called by parent groups in configure.

        Parameters
        ----------
        name : str
            name of the variable in this component's namespace.
        shape : int or tuple or list or None
            Shape of this variable, only required if val is not an array.
            Default is None.
        units : str or None
            Units in which the output variables will be provided to the component during execution.
            Default is None, which means it has no units.
        desc : str
            description of the timeseries output variable.
        src : str
            The src path of the variables input, used to prevent redundant inputs.
        rate : bool
            If True, timeseries output is a rate.

        Returns
        -------
        bool
            True if a new input was added for the output, or False if it reuses an existing input.
        """
        input_num_nodes = self.input_num_nodes
        output_num_nodes = self.output_num_nodes
        added_source = False

        if name in self._vars:
            return False

        if src in self._sources:
            # If we're already pulling the source into this timeseries, use that as the
            # input for this output.
            input_name = self._sources[src]
            input_units = self._units[input_name]
        else:
            input_name = f'input_values:{name}'
            self.add_input(input_name,
                           shape=(input_num_nodes,) + shape,
                           units=units, desc=desc)
            self._sources[src] = input_name
            input_units = self._units[input_name] = units
            added_source = True

        output_name = name
        self.add_output(output_name, shape=(output_num_nodes,) + shape, units=units, desc=desc)

        self._vars[name] = (input_name, output_name, shape, rate)

        size = np.prod(shape)

        # Get any scaling that needs to be performed for the derivative
        if input_units is None or units is None:
            scale = 1.0
        else:
            scale, offset = unit_conversion(input_units, units)
            self._conversion_factors[output_name] = scale, offset

        if not rate and self._no_interp:
            # Case 1: No outputs just echo inputs.
            # Jacobian values are just the scale factor
            jac_rows = jac_cols = np.arange(input_num_nodes * size, dtype=int)
            jac_val = scale
        else:
            if rate:
                # Case 2: The output is a rate
                # Jacobian values depend on the differentiation matrix
                # This is not well conditioned for large numbers of nodes.
                mat = self.differentiation_matrix
            else:
                # Case 3: the output is an interpolation
                # Jacobian values depend on the interpolation matrix.
                # This is subject to noise with large numbers of nodes.
                mat = self.interpolation_matrix

            # Preallocate lists to hold data for jac
            jac_data = []
            jac_indices = []
            jac_indptr = [0]

            # Iterate over the dense dimension 'size'
            for s in range(size):
                # Extend the data and indices using the CSR attributes of mat
                jac_data.extend(mat.data)
                jac_indices.extend(mat.indices + s * input_num_nodes)

                # For every non-zero row in mat, update jac's indptr
                new_indptr = mat.indptr[1:] + s * len(mat.data)
                jac_indptr.extend(new_indptr)

            # Correct the last entry of jac_indptr
            jac_indptr[-1] = len(jac_data)

            # Construct the sparse jac matrix in CSR format
            jac = sp.csr_matrix((jac_data, jac_indices, jac_indptr),
                                shape=(output_num_nodes * size, input_num_nodes * size))

            # Now, if you want to get the row and column indices of the non-zero entries in the jac matrix:
            jac_rows, jac_cols = jac.nonzero()

            # There's a chance that the input for this output was pulled from another variable with
            # different units, so account for that with a conversion.
            jac_val = np.squeeze(np.array(jac[jac_rows, jac_cols])) * scale

        self.declare_partials(of=output_name, wrt=input_name,
                              rows=jac_rows, cols=jac_cols,
                              val=jac_val)

        return added_source

    def _build_matrices(self, compute_L=True, compute_D=True):
        """
        Build the interpolation and differentiation matrices for the component, if necessary.

        Parameters
        ----------
        compute_L : bool, optional
            If True, compute and store the interpolation matrix. Otherwise return None in its place. Default is True.
        compute_D : bool, optional
            If True, compute and store the diferentiation matrix. Otherwise return None in its place. Default is True.
        """
        igd = self.options['input_grid_data']
        ogd = self.options['output_grid_data']
        output_subset = self.options['output_subset']

        if ogd is None:
            ogd = igd

        self.input_num_nodes = igd.num_nodes
        self.output_num_nodes = ogd.subset_num_nodes[output_subset]

        self.add_input('dt_dstau', shape=(self.input_num_nodes,), units=self.options['time_units'])

        # Build the interpolation matrix which maps from the input grid to the output grid.
        # Rather than a single phase-wide interpolating polynomial, map each segment.
        # To do this, find the nodes in the output grid which fall in each segment of the input
        # grid.  Then build a Lagrange interpolating polynomial for that segment
        L_blocks = []
        D_blocks = []
        output_nodes_ptau = ogd.node_ptau[ogd.subset_node_indices[output_subset]]

        for iseg in range(igd.num_segments):
            i1, i2 = igd.segment_indices[iseg]
            iptau_segi = igd.node_ptau[i1:i2]
            istau_segi = igd.node_stau[i1:i2]

            # The indices of the output grid that fall within this segment of the input grid
            if ogd is igd and output_subset == 'all':
                optau_segi = iptau_segi
            else:
                ptau_hi = igd.segment_ends[iseg+1]
                if iseg < igd.num_segments - 1:
                    optau_segi = output_nodes_ptau[output_nodes_ptau <= ptau_hi]
                else:
                    optau_segi = output_nodes_ptau

                # Remove the captured nodes so we don't accidentally include them again
                output_nodes_ptau = output_nodes_ptau[len(optau_segi):]

            # # Now get the output nodes which fall in iseg in iseg's segment tau space.
            ostau_segi = 2.0 * (optau_segi - iptau_segi[0]) / (iptau_segi[-1] - iptau_segi[0]) - 1

            # Create the interpolation matrix and add it to the blocks
            L, D = lagrange_matrices(istau_segi, ostau_segi,
                                     compute_interp_matrix=compute_L,
                                     compute_diff_matrix=compute_D)
            L_blocks.append(L)
            D_blocks.append(D)

        if compute_L:
            self.interpolation_matrix = sp.block_diag(L_blocks, format='csr')
        if compute_D:
            self.differentiation_matrix = sp.block_diag(D_blocks, format='csr')

    def _configure_io(self, timeseries_options):
        """
        Configure the component IO for the given timeseries.

        Parameters
        ----------
        phase : Phase
            The phase to which this component belongs.
        timeseries_name : str
            The name of the timeseries being configured.

        Returns
        -------
        inputs : list
            The timeseries inputs to this component.
        """
        igd = self.options['input_grid_data']
        ogd = self.options['output_grid_data']
        output_subset = self.options['output_subset']

        if ogd is None:
            ogd = igd

        if ogd == igd and output_subset == 'all':
            self._no_interp = True

        self.input_num_nodes = igd.num_nodes
        self.output_num_nodes = ogd.subset_num_nodes[output_subset]

        # Check if we ever use rates, if not, don't bother computing/storing [D]
        compute_D = any([ts_options['is_rate'] for ts_options in timeseries_options['outputs'].values()])
        compute_L = not self._no_interp

        # Compute and store the interpolation and differentiation matrices, if necessary.
        self._build_matrices(compute_L, compute_D)

        # Add the IO for the given timeseries
        ts_inputs = []
        for ts_output_name, ts_output in timeseries_options['outputs'].items():
            name = ts_output['output_name'] if ts_output['output_name'] is not None else ts_output['name']
            units = ts_output['units']
            shape = ts_output['shape']
            src = ts_output['src']
            is_rate = ts_output['is_rate']

            added_src = self._add_output_configure(name, shape=shape,
                                                   units=units, desc='',
                                                   src=src, rate=is_rate)

            if added_src:
                # If we hadn't already used this input, add it now.
                ts_inputs.append((f'input_values:{name}', src, ts_output['src_idxs']))
        return ts_inputs

    def compute(self, inputs, outputs):
        """
        Compute component outputs.

        Parameters
        ----------
        inputs : `Vector`
            `Vector` containing inputs.
        outputs : `Vector`
            `Vector` containing outputs.
        """
        # convert dt_dstau to a column vector
        dt_dstau = inputs['dt_dstau'][:, np.newaxis]

        for (input_name, output_name, _, is_rate) in self._vars.values():
            if self._no_interp:
                interp_vals = inputs[input_name]
            else:
                inp = inputs[input_name]
                if len(inp.shape) > 2:
                    # Dot product always performs the sum product over axis 2.
                    inp = inp.swapaxes(0, 1)
                interp_vals = self.interpolation_matrix.dot(inp)

            if is_rate:
                interp_vals = self.differentiation_matrix.dot(interp_vals) / dt_dstau

            if output_name in self._conversion_factors:
                scale, offset = self._conversion_factors[output_name]
                outputs[output_name] = scale * (interp_vals + offset)
            else:
                outputs[output_name] = interp_vals



from __future__ import print_function, division, absolute_import
import unittest

from openmdao.api import Problem, Group, pyOptSparseDriver
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse

import dymos as dm
from dymos import Trajectory, GaussLobatto, Phase, Radau, ExplicitShooting
from dymos.examples.hull_problem.hull_ode import HullProblemODE

c = 5


@use_tempdirs
class TestHull(unittest.TestCase):

    @staticmethod
    @require_pyoptsparse(optimizer='IPOPT')
    def make_problem(transcription=GaussLobatto, numseg=30):
        p = Problem(model=Group())
        p.driver = pyOptSparseDriver()
        p.driver.declare_coloring()
        p.driver.options['optimizer'] = 'IPOPT'
        p.driver.opt_settings['hessian_approximation'] = 'limited-memory'
        p.driver.opt_settings['print_level'] = 0
        p.driver.opt_settings['linear_solver'] = 'mumps'
        p.driver.declare_coloring()

        traj = p.model.add_subsystem('traj', Trajectory())
        phase0 = traj.add_phase('phase', Phase(ode_class=HullProblemODE,
                                               transcription=transcription(num_segments=numseg, order=3)))
        phase0.set_time_options(fix_initial=True, fix_duration=True)
        phase0.add_state('x', fix_initial=True, fix_final=False, rate_source='u')
        phase0.add_state('xL', fix_initial=True, fix_final=False, rate_source='L')
        phase0.add_control('u', opt=True, targets=['u'])

        phase0.add_objective(f'J = {c}*x**2/2 + xL')

        p.setup(check=True)

        phase0.set_state_val('x', [1.5, 1])
        phase0.set_state_val('xL', [0, 1])
        phase0.set_time_val(initial=0.0, duration=10.0)
        phase0.set_control_val('u', [-7, -0.14])
        return p

    @staticmethod
    def solution(x0, td):

        xf = x0 - c * x0 * td / (1 + c * td)
        uf = -c * x0 / (1 + c * td)

        return xf, uf

    def test_hull_gauss_lobatto(self):
        p = self.make_problem(transcription=GaussLobatto)
        dm.run_problem(p, simulate=True)

        xf, uf = self.solution(1.5, 10)

        assert_near_equal(p.get_val('traj.phase.timeseries.x')[-1],
                          xf,
                          tolerance=1e-4)

        assert_near_equal(p.get_val('traj.phase.timeseries.u')[-1],
                          uf,
                          tolerance=1e-4)

    def test_hull_radau(self):
        p = self.make_problem(transcription=Radau)
        dm.run_problem(p, simulate=True)

        xf, uf = self.solution(1.5, 10)

        assert_near_equal(p.get_val('traj.phase.timeseries.x')[-1],
                          xf,
                          tolerance=1e-4)

        assert_near_equal(p.get_val('traj.phase.timeseries.u')[-1],
                          uf,
                          tolerance=1e-4)

    def test_hull_shooting(self):
        p = self.make_problem(transcription=ExplicitShooting)
        dm.run_problem(p, simulate=True)

        xf, uf = self.solution(1.5, 10)

        assert_near_equal(p.get_val('traj.phase.timeseries.x')[-1],
                          xf,
                          tolerance=1e-4)

        assert_near_equal(p.get_val('traj.phase.timeseries.u')[-1],
                          uf,
                          tolerance=1e-4)



import numpy as np

import openmdao.api as om


class MassComp(om.ExplicitComponent):
    """ Compute the total mass of the aircraft """

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']
        self.add_input('mass_fuel', val=np.ones(nn), desc='fuel mass', units='kg')
        self.add_input('mass_empty', val=np.ones(nn), desc='empty aircraft mass', units='kg')
        self.add_input('mass_payload', val=np.ones(nn), desc='aircraft payload mass', units='kg')

        self.add_output('mass_total', val=np.ones(nn), desc='total aircraft mass', units='kg')
        self.add_output('W_total', val=np.ones(nn), desc='total aircraft weight', units='N')

        # Setup partials
        ar = np.arange(self.options['num_nodes'])

        self.declare_partials(of='mass_total', wrt='mass_fuel', rows=ar, cols=ar, val=1.0)
        self.declare_partials(of='mass_total', wrt='mass_empty', rows=ar, cols=ar, val=1.0)
        self.declare_partials(of='mass_total', wrt='mass_payload', rows=ar, cols=ar, val=1.0)

        self.declare_partials(of='W_total', wrt='mass_fuel', rows=ar, cols=ar, val=9.80665)
        self.declare_partials(of='W_total', wrt='mass_empty', rows=ar, cols=ar, val=9.80665)
        self.declare_partials(of='W_total', wrt='mass_payload', rows=ar, cols=ar, val=9.80665)

    def compute(self, inputs, outputs):
        outputs['mass_total'] = inputs['mass_fuel'] + inputs['mass_empty'] + inputs['mass_payload']
        outputs['W_total'] = 9.80665 * outputs['mass_total']



import openmdao.api as om
import numpy as np

from .tracks import ovaltrack
from .spline import get_track_points, get_spline


class Curvature(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=int)
        track = ovaltrack  # remember to change here and in problemSolver.py

        points = get_track_points(track)
        finespline, gates, gatesd, curv, slope = get_spline(points)

        self.curv = curv
        self.track_length = track.get_total_length()

    def setup(self):
        nn = self.options['num_nodes']

        # constants
        self.add_input('s', val=np.zeros(nn), desc='distance along track', units='m')

        # outputs
        self.add_output('kappa', val=np.zeros(nn), desc='track centerline Curvature', units='1/m')

        # no partials needed

    def compute(self, inputs, outputs):
        s = inputs['s']

        num_curv_points = len(self.curv)

        kappa = np.zeros(len(s))

        for i in range(len(s)):
            index = np.floor((s[i]/self.track_length)*num_curv_points)
            index = np.minimum(index, num_curv_points-1)
            kappa[i] = self.curv[index.astype(int)]

        outputs['kappa'] = kappa



import numpy as np

import openmdao.api as om
from ..transcription_base import TranscriptionBase
from ..common import TimeComp
from .components import StateIndependentsComp, StateInterpComp, CollocationComp
from ..common.timeseries_output_comp import TimeseriesOutputComp
from ...utils.misc import CoerceDesvar, get_rate_units, reshape_val
from ...utils.introspection import get_promoted_vars, get_source_metadata
from ...utils.constants import INF_BOUND
from ...utils.indexing import get_src_indices_by_row


class PseudospectralBase(TranscriptionBase):
    """
    Base class for the pseudospectral transcriptions.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.any_solved_segs = False
        self.any_connected_opt_segs = False

    def initialize(self):
        """
        Declare transcription options.
        """
        self.options.declare(name='solve_segments', default=False,
                             values=(False, 'forward', 'backward'),
                             desc='Applies \'solve_segments\' behavior to _all_ states in the Phase. '
                                  'If \'forward\', collocation defects within each '
                                  'segment are solved with a Newton solver by fixing the initial value in the '
                                  'phase (if using compressed transcription) or segment (if not using '
                                  'compressed transcription). This provides a forward shooting (or multiple shooting) '
                                  'method.  If \'backward\', the final value in the phase or segment is fixed '
                                  'and a solver finds the other ones to mimic reverse propagation. Set '
                                  'to False (the default) to explicitly disable the use of a solver to '
                                  'converge the state time history.')

    def setup_time(self, phase):
        """
        Setup the time component.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        grid_data = self.grid_data

        super(PseudospectralBase, self).setup_time(phase)

        time_comp = TimeComp(num_nodes=grid_data.num_nodes, node_ptau=grid_data.node_ptau,
                             node_dptau_dstau=grid_data.node_dptau_dstau,
                             units=phase.time_options['units'],
                             initial_val=phase.time_options['initial_val'],
                             duration_val=phase.time_options['duration_val'])

        phase.add_subsystem('time', time_comp, promotes_inputs=['*'], promotes_outputs=['*'])

    def configure_time(self, phase):
        """
        Configure the inputs/outputs on the time component.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super(PseudospectralBase, self).configure_time(phase)
        phase.time.configure_io()

    def setup_states(self, phase):
        """
        Setup the states for this transcription.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        grid_data = self.grid_data

        for options in phase.state_options.values():
            # Transcription solve_segments overrides state solve_segments if its not set
            if options['solve_segments'] is None:
                options['solve_segments'] = self.options['solve_segments']

            if options['solve_segments']:
                self.any_solved_segs = True
            elif options['input_initial']:
                self.any_connected_opt_segs = True

        if self.any_solved_segs or self.any_connected_opt_segs:
            indep = StateIndependentsComp(grid_data=grid_data,
                                          state_options=phase.state_options)
        else:
            indep = om.IndepVarComp()

        num_connected = len([s for (s, opts) in phase.state_options.items() if opts['input_initial']])
        prom_inputs = ['initial_states:*'] if num_connected > 0 else None
        phase.add_subsystem('indep_states', indep, promotes_inputs=prom_inputs,
                            promotes_outputs=['*'])

    def configure_controls(self, phase):
        """
        Configure control I/O for the phase.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super().configure_controls(phase)

        if phase.control_options:
            phase.control_comp.configure_io()
            phase.promotes('control_comp',
                           any=['*controls:*', '*control_values:*', '*control_rates:*'])

            phase.connect('dt_dstau', 'control_comp.dt_dstau')

    def configure_states(self, phase):
        """
        Configure state connections post-introspection.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super().configure_states(phase)
        grid_data = self.grid_data
        num_state_input_nodes = grid_data.subset_num_nodes['state_input']
        indep = phase.indep_states

        # state_idx_map holds the node indices provided by the solver (solver) and those
        # that are independent variables (indep)
        self.state_idx_map = {}

        # add all the des-vars (either from the IndepVarComp or from the indep-var-like
        # outputs of the collocation comp)
        for name, options in phase.state_options.items():
            self._configure_solve_segments(name, options, phase)
            shape = options['shape']
            # In certain cases, we put an output on the IVC.
            if isinstance(indep, om.IndepVarComp):
                default_val = reshape_val(options['val'], shape, num_state_input_nodes)
                indep.add_output(name=f'states:{name}',
                                 shape=(num_state_input_nodes,) + shape,
                                 val=default_val,
                                 units=options['units'])

            if options['opt']:
                # Add the states as design variables.
                #
                # In the case of optimizer-driven collocation, this includes the values at all
                # nodes (excluding initial and/or final if fix_initial and/or fix_final is specified).
                #
                # In the case of solve_segments == 'forward', the design var nodes are the first
                # node in the  phase (when compressed) or the first node in each each segment
                # (when not compressed).
                #
                # When fix_initial is True, the first node in the phase is then removed from the
                # design variable node indices. (So when compressed, there is no design variable).
                #
                # In the case of solve_segments == 'backward', the design var nodes are the last
                # node in the phase (when compressed) or the last node in each segment (when not
                # compressed).
                #
                # When fix_final is True, the last node in the phase is then removed from the design
                # variable node indices.
                #
                desvar_node_idxs = np.asarray(self.state_idx_map[name]['indep'])

                # This matrix will contain 1's for every index that is to be a design variable,
                # otherwise the value will be zero
                state_input_shape = (num_state_input_nodes,) + options['shape']
                idx_mask = np.zeros(state_input_shape, dtype=int)
                idx_mask[desvar_node_idxs, ...] = 1

                if options['fix_initial']:
                    if options['initial_bounds'] is not None:
                        raise ValueError('Cannot specify \'fix_initial=True\' and specify '
                                         f'initial_bounds for state {name} in phase {phase.name}')
                    if options['input_initial']:
                        raise ValueError('Cannot specify \'fix_initial=True\' and specify '
                                         f'\'connected_initial=True\' for state {name} '
                                         f'in phase {phase.name}')
                    idx_mask[0, ...] = np.asarray(np.logical_not(options['fix_initial']), dtype=int)
                elif options['input_initial']:
                    if options['initial_bounds'] is not None:
                        raise ValueError('Cannot specify \'connected_initial=True\' and specify '
                                         f'initial_bounds for state {name} in phase {phase.name}')
                    idx_mask[0, ...] = np.asarray(np.logical_not(options['input_initial']), dtype=int)

                if options['fix_final']:
                    if options['final_bounds'] is not None:
                        raise ValueError('Cannot specify \'fix_final=True\' and specify '
                                         f'final_bounds for state {name}')
                    idx_mask[-1, ...] = np.asarray(np.logical_not(options['fix_final']), dtype=int)

                # Now convert the masked array into actual flat indices
                desvar_indices = np.arange(idx_mask.size, dtype=int).reshape(state_input_shape)[idx_mask.nonzero()]

                if len(desvar_indices) > 0:
                    coerce_desvar_option = CoerceDesvar(num_state_input_nodes, desvar_indices,
                                                        options)

                    lb = np.zeros_like(desvar_indices, dtype=float)
                    lb[:] = -INF_BOUND if coerce_desvar_option('lower') is None else \
                        coerce_desvar_option('lower')

                    ub = np.zeros_like(desvar_indices, dtype=float)
                    ub[:] = INF_BOUND if coerce_desvar_option('upper') is None else \
                        coerce_desvar_option('upper')

                    if options['initial_bounds'] is not None:
                        lb[0] = options['initial_bounds'][0]
                        ub[0] = options['initial_bounds'][-1]

                    if options['final_bounds'] is not None:
                        lb[-1] = options['final_bounds'][0]
                        ub[-1] = options['final_bounds'][-1]

                    phase.add_design_var(name=f'states:{name}',
                                         lower=lb,
                                         upper=ub,
                                         scaler=coerce_desvar_option('scaler'),
                                         adder=coerce_desvar_option('adder'),
                                         ref0=coerce_desvar_option('ref0'),
                                         ref=coerce_desvar_option('ref'),
                                         indices=desvar_indices,
                                         flat_indices=True)

        if isinstance(indep, StateIndependentsComp):
            indep.configure_io(self.state_idx_map)

        if self.any_solved_segs or self.any_connected_opt_segs:
            for name, options in phase.state_options.items():
                if options['solve_segments']:
                    phase.connect(f'collocation_constraint.defects:{name}',
                                  f'indep_states.defects:{name}')

    def setup_ode(self, phase):
        """
        Setup the ode for this transcription.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        grid_data = self.grid_data
        phase.add_subsystem('state_interp',
                            subsys=StateInterpComp(grid_data=grid_data,
                                                   state_options=phase.state_options,
                                                   time_units=phase.time_options['units'],
                                                   transcription=grid_data.transcription))

    def configure_ode(self, phase):
        """
        Create connections to the introspected states.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        grid_data = self.grid_data
        map_input_indices_to_disc = grid_data.input_maps['state_input_to_disc']

        phase.state_interp.configure_io()

        phase.connect('dt_dstau', 'state_interp.dt_dstau',
                      src_indices=grid_data.subset_node_indices['col'], flat_src_indices=True)

        for name in phase.state_options:
            phase.connect(f'states:{name}',
                          f'state_interp.state_disc:{name}',
                          src_indices=om.slicer[map_input_indices_to_disc, ...])

    def setup_defects(self, phase):
        """
        Setup the Collocation and Continuity components as necessary.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        phase.add_subsystem('collocation_constraint',
                            CollocationComp(grid_data=self.grid_data,
                                            state_options=phase.state_options,
                                            time_units=phase.time_options['units']))

    def _configure_solve_segments(self, state_name, options, phase):
        """
        Provides error checking for solve_segments and establishes necessary data structures.

        Parameters
        ----------
        state_name : str
            The name of the state being configured.
        options : StateOptionsDictionary
            The StateOptionsDictionary for the state being configured.
        phase : Phase
            The Dymos Phase associated with this transcription instance.
        """
        self.state_idx_map[state_name] = {'solver': None, 'indep': None}

        state_input_idxs = self.grid_data.subset_node_indices['state_input']
        num_state_input_nodes = self.grid_data.subset_num_nodes['state_input']
        compressed = self.options['compressed']

        # Sanity-checks for solve segments
        # If solve_segments is used at all, we cannot fix the state at both ends of the phase.
        if options['solve_segments']:
            if options['fix_initial'] and options['fix_final']:
                raise ValueError(f'Can not use solve_segments for state ({state_name}) '
                                 f'in phase ({phase.name}) with both "fix_initial" and '
                                 '"fix_final" set to True.')
            if options['input_initial'] and options['fix_final']:
                raise ValueError(f'Can not use solve_segments for state ({state_name}) '
                                 f'in phase ({phase.name}) with both "input_initial" '
                                 f'and "fix_final" set to True .')

            # If solve_segments is 'forward', 'fix_final' may not be True
            if options['solve_segments'] in {'forward', True}:
                if options['fix_final']:
                    raise ValueError(f'Cannot use solve_segments in phase ({phase.name}) for state '
                                     f'({state_name}) with forward propagation when fix_final=True.'
                                     f' Either set fix_final=False or set solve_segments=\'backward\'')

            # Backward propagation
            if options['solve_segments'] == 'backward':
                # Neither 'fix_initial' nor 'input_initial' may be True.
                if options['fix_initial']:
                    raise ValueError(f'Cannot use solve_segments in phase ({phase.name}) with '
                                     f'backward propagation when fix_initial=True. Either set '
                                     f'fix_final=False or set solve_segments=\'backward\'')
                elif options['input_initial']:
                    raise ValueError(f'Cannot use solve_segments in phase ({phase.name}) with '
                                     f'backward propagation when input_initial=True. Either set '
                                     f'input_initial=False or set solve_segments=\'forward\'')

                if compressed:
                    # The optimizer controls the last state input node, all others are solver-controlled
                    self.state_idx_map[state_name]['indep'] = np.array([num_state_input_nodes - 1], dtype=int)
                    self.state_idx_map[state_name]['solver'] = np.arange(num_state_input_nodes - 1, dtype=int)
                else:
                    # The optimizer controls the last state input node in each segment, all others are solver_controlled
                    right_idxs = self.grid_data.subset_node_indices['segment_ends'][1::2]
                    self.state_idx_map[state_name]['indep'] = [i for i in range(num_state_input_nodes)
                                                               if state_input_idxs[i] in right_idxs]
                    self.state_idx_map[state_name]['solver'] = [i for i in range(num_state_input_nodes)
                                                                if state_input_idxs[i] not in right_idxs]
            # Forward propagation
            elif options['solve_segments'] in {'forward'}:
                if compressed:
                    self.state_idx_map[state_name]['solver'] = np.arange(1, num_state_input_nodes, dtype=int)
                    self.state_idx_map[state_name]['indep'] = np.zeros((1,), dtype=int)
                else:
                    left_idxs = self.grid_data.subset_node_indices['segment_ends'][0::2]
                    self.state_idx_map[state_name]['solver'] = [i for i in range(num_state_input_nodes)
                                                                if state_input_idxs[i] not in left_idxs]
                    self.state_idx_map[state_name]['indep'] = [i for i in range(num_state_input_nodes)
                                                               if state_input_idxs[i] in left_idxs]
        else:
            # No solver used to solve these nodes.  All state input nodes are the indep nodes.
            self.state_idx_map[state_name]['solver'] = []
            self.state_idx_map[state_name]['indep'] = np.arange(len(state_input_idxs), dtype=int)

    def configure_defects(self, phase):
        """
        Configure the continuity_comp and connect the collocation constraints.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        grid_data = self.grid_data

        phase.connect('dt_dstau', ('collocation_constraint.dt_dstau'),
                      src_indices=grid_data.subset_node_indices['col'], flat_src_indices=True)

        phase.collocation_constraint.configure_io()

        any_state_cnty, any_control_cnty, any_control_rate_cnty = self._requires_continuity_constraints(phase)

        if not any((any_state_cnty, any_control_cnty, any_control_rate_cnty)):
            return

        # Add the continuity constraint component if necessary
        segment_end_idxs = grid_data.subset_node_indices['segment_ends']
        state_disc_idxs = grid_data.subset_node_indices['state_disc']

        if any_state_cnty:
            state_input_subidxs = np.where(np.isin(state_disc_idxs, segment_end_idxs))[0]

            for name, options in phase.state_options.items():
                shape = options['shape']
                flattened_src_idxs = get_src_indices_by_row(state_input_subidxs, shape=shape,
                                                            flat=True)
                phase.connect(f'states:{name}',
                              f'continuity_comp.states:{name}',
                              src_indices=(flattened_src_idxs,), flat_src_indices=True)

        if any_control_rate_cnty:
            phase.connect('t_duration_val', 'continuity_comp.t_duration')

        for name, options in phase.control_options.items():
            # The sub-indices of control_disc indices that are segment ends
            segment_end_idxs = grid_data.subset_node_indices['segment_ends']
            src_idxs = get_src_indices_by_row(segment_end_idxs, options['shape'], flat=True)

            # enclose indices in tuple to ensure shaping of indices works
            src_idxs = (src_idxs,)

            if options['continuity']:
                phase.connect(f'control_values:{name}',
                              f'continuity_comp.controls:{name}',
                              src_indices=src_idxs, flat_src_indices=True)

            if options['rate_continuity']:
                phase.connect(f'control_rates:{name}_rate',
                              f'continuity_comp.control_rates:{name}_rate',
                              src_indices=src_idxs, flat_src_indices=True)

            if options['rate2_continuity']:
                phase.connect(f'control_rates:{name}_rate2',
                              f'continuity_comp.control_rates:{name}_rate2',
                              src_indices=src_idxs, flat_src_indices=True)

    def setup_solvers(self, phase):
        """
        Setup the solvers.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        pass

    def configure_solvers(self, phase, requires_solvers=None):
        """
        Configure the solvers.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        requires_solvers : dict[str: bool]
            A dictionary mapping a string descriptor of a reason why a solver is required,
            and whether a solver is required.
        """
        req_solvers = {'solved segments': self.any_solved_segs,
                       'input initial': self.any_connected_opt_segs}
        if requires_solvers is not None:
            req_solvers.update(requires_solvers)
        super().configure_solvers(phase, requires_solvers=req_solvers)

    def setup_timeseries_outputs(self, phase):
        """
        Setup the timeseries for this transcription.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        gd = self.grid_data

        for name, options in phase._timeseries.items():
            if options['transcription'] is None:
                ogd = None
            else:
                ogd = options['transcription'].grid_data

            timeseries_comp = TimeseriesOutputComp(input_grid_data=gd,
                                                   output_grid_data=ogd,
                                                   output_subset=options['subset'],
                                                   time_units=phase.time_options['units'])
            phase.add_subsystem(name, subsys=timeseries_comp)

            phase.connect('dt_dstau', f'{name}.dt_dstau', flat_src_indices=True)

    def _get_response_src(self, var, loc, phase, ode_outputs=None):
        """
        Return the path to the variable that will be used as a response..

        Parameters
        ----------
        var : str
            Name of the variable to be used as the response.
        loc : str
            The location of the response in the phase ['initial', 'final'].
        phase : dymos.Phase
            Phase object containing in which the objective resides.
        ode_outputs : dict or None
            A dictionary of ODE outputs as returned by get_promoted_vars.

        Returns
        -------
        obj_path : str
            Path to the source.
        shape : tuple
            Source shape.
        units : str
            Source units.
        linear : bool
            True if the objective quantity is linear.
        """
        time_units = phase.time_options['units']
        var_type = phase.classify_var(var)

        if ode_outputs is None:
            ode_outputs = get_promoted_vars(phase._get_subsystem(self._rhs_source), 'output')

        if var_type == 't':
            shape = (1,)
            units = time_units
            linear = True
            constraint_path = 't'
        elif var_type == 't_phase':
            shape = (1,)
            units = time_units
            linear = True
            constraint_path = 't_phase'
        elif var_type == 'state':
            shape = phase.state_options[var]['shape']
            units = phase.state_options[var]['units']
            solve_segments = phase.state_options[var]['solve_segments']
            connected_initial = phase.state_options[var]['input_initial']
            if not solve_segments and not connected_initial:
                linear = True
            elif solve_segments in {'forward'} and not connected_initial and loc == 'initial':
                linear = True
            elif solve_segments == 'backward' and loc == 'final':
                linear = True
            else:
                linear = False
            constraint_path = f'states:{var}'
        elif var_type == 'control':
            shape = phase.control_options[var]['shape']
            units = phase.control_options[var]['units']
            linear = False
            constraint_path = f'control_values:{var}'
        elif var_type == 'parameter':
            shape = phase.parameter_options[var]['shape']
            units = phase.parameter_options[var]['units']
            linear = True
            constraint_path = f'parameter_vals:{var}'
        elif var_type in ('control_rate', 'control_rate2'):
            control_var = var[:-5] if var_type == 'control_rate' else var[:-6]
            shape = phase.control_options[control_var]['shape']
            control_units = phase.control_options[control_var]['units']
            d = 2 if var_type == 'control_rate2' else 1
            control_rate_units = get_rate_units(control_units, time_units, deriv=d)
            units = control_rate_units
            linear = False
            constraint_path = f'control_rates:{var}'
        else:
            # Failed to find variable, assume it is in the ODE. This requires introspection.
            constraint_path = f'{self._rhs_source}.{var}'
            meta = get_source_metadata(ode_outputs, var, user_units=None, user_shape=None)
            shape = meta['shape']
            units = meta['units']
            linear = False

        return constraint_path, shape, units, linear

    def _get_num_timeseries_nodes(self):
        """
        Returns the number of nodes in the default timeseries for this transcription.

        Returns
        -------
        int
            The number of nodes in the default timeseries for this transcription.
        """
        return self.grid_data.num_nodes

    def _phase_set_state_val(self, phase, name, vals, time_vals, interpolation_kind):
        """
        Method to interpolate the provided input and return the variables that need to be set
        along with their appropriate value.

        Parameters
        ----------
        phase : dymos.Phase
            The phase to which this transcription applies.
        name : str
            The name of the phase variable to be set.
        vals : ndarray or Sequence or float
            Array of control/state/parameter values.
        time_vals : ndarray or Sequence or None
            Array of integration variable values.
        interpolation_kind : str
            Specifies the kind of interpolation, as per the scipy.interpolate package.
            One of ('linear', 'nearest', 'zero', 'slinear', 'quadratic', 'cubic'
            where 'zero', 'slinear', 'quadratic' and 'cubic' refer to a spline
            interpolation of zeroth, first, second or third order) or as an
            integer specifying the order of the spline interpolator to use.
            Default is 'linear'.

        Returns
        -------
        input_data : dict
            Dict containing the values that need to be set in the phase

        """

        input_data = {}
        if np.isscalar(vals):
            interp_vals = vals
        else:
            interp_vals = phase.interp(name, ys=vals, xs=time_vals,
                                       nodes='state_input',
                                       kind=interpolation_kind)
        input_data[f'states:{name}'] = interp_vals

        return input_data



import unittest

import numpy as np
from numpy.testing import assert_almost_equal

import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials

from dymos.examples.min_time_climb.aero import AeroGroup


class TestAeroGroup(unittest.TestCase):

    def setUp(self):
        self.prob = om.Problem(model=om.Group())
        nn = 5

        ivc = om.IndepVarComp()

        ivc.add_output('rho', val=0.0001 * np.ones(nn), units='kg/m**3')
        ivc.add_output('v', val=0.0001 * np.ones(nn), units='m/s')
        ivc.add_output('S', val=0.0001 * np.ones(nn), units='m**2')
        ivc.add_output('alpha', val=np.zeros(nn), units='rad')
        ivc.add_output('sos', val=np.ones(nn), units='m/s')
        self.prob.model.add_subsystem(name='ivc', subsys=ivc, promotes_outputs=['*'])

        self.prob.model.add_subsystem(name='aero', subsys=AeroGroup(num_nodes=nn))

        self.prob.model.connect('rho', 'aero.rho')
        self.prob.model.connect('v', 'aero.v')
        self.prob.model.connect('S', 'aero.S')
        self.prob.model.connect('alpha', 'aero.alpha')
        self.prob.model.connect('sos', 'aero.sos')

        self.prob.setup(force_alloc_complex=True)

    def test_aero_values(self):

        self.prob['rho'] = 1.2250409
        self.prob['v'] = 115.824
        self.prob['S'] = 49.2386

        self.prob['alpha'] = np.radians(4.1900199)
        self.prob['sos'] = 340.29396
        self.prob.run_model()

        q_expected = 0.5 * self.prob['rho'] * self.prob['v']**2
        lift_expected = q_expected * self.prob['S'] * self.prob['aero.CL']   # [101779.451502]
        drag_expected = q_expected * self.prob['S'] * self.prob['aero.CD']   # [9278.85725577]

        assert_almost_equal(self.prob['aero.q'], q_expected, decimal=7)
        assert_almost_equal(self.prob['aero.f_lift'], lift_expected, decimal=7)
        assert_almost_equal(self.prob['aero.f_drag'], drag_expected, decimal=7)

    def testAeroDerivs(self):
        cpd = self.prob.check_partials(compact_print=True, method='cs', out_stream=None)
        assert_check_partials(cpd, atol=1.0E-5, rtol=1.0E-5)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest

import openmdao.api as om
import dymos as dm

from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs
from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE


@use_tempdirs
class TestBrachistochroneBirkhoffConstraints(unittest.TestCase):

    def test_brachistochrone_control_prefix(self):

        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.options['optimizer'] = 'SLSQP'
        p.driver.declare_coloring(tol=1.0E-12)

        tx = dm.Birkhoff(num_nodes=25, grid_type='lgl')

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)
        phase.timeseries_options['use_prefix'] = True
        p.model.add_subsystem('traj', traj)
        traj.add_phase('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=False)
        phase.add_state('y', fix_initial=True, fix_final=False)

        # Note that by omitting the targets here Dymos will automatically attempt to connect
        # to a top-level input named 'v' in the ODE, and connect to nothing if it's not found.
        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta',
                          continuity=True, rate_continuity=True,
                          units='deg')

        phase.add_parameter('g', targets=['g'], units='m/s**2')

        phase.add_path_constraint('theta', lower=0.01, upper=179.9)
        phase.add_boundary_constraint('theta', loc='final', lower=0.01, upper=179.9)

        phase.add_boundary_constraint('x', loc='final', equals=10)
        phase.add_boundary_constraint('y', loc='final', equals=5)
        # Minimize time at the end of the phase
        phase.add_objective('time_phase', loc='final', scaler=10)

        p.set_solver_print(0)

        p.setup()

        phase.set_time_val(initial=0.0, duration=1.5)

        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])

        phase.set_control_val('theta', [5, 100])
        phase.set_parameter_val('g', 9.80665)

        p.run_driver()
        assert_near_equal(p.get_val('traj.phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-4)

    def test_brachistochrone_control_no_prefix(self):

        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.options['optimizer'] = 'SLSQP'
        p.driver.declare_coloring(tol=1.0E-12)

        tx = dm.Birkhoff(num_nodes=25, grid_type='lgl')

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)
        phase.timeseries_options['use_prefix'] = False
        p.model.add_subsystem('traj', traj)
        traj.add_phase('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=False)
        phase.add_state('y', fix_initial=True, fix_final=False)

        # Note that by omitting the targets here Dymos will automatically attempt to connect
        # to a top-level input named 'v' in the ODE, and connect to nothing if it's not found.
        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta',
                          continuity=True, rate_continuity=True,
                          units='deg')

        phase.add_parameter('g', targets=['g'], units='m/s**2')

        phase.add_path_constraint('theta', lower=0.01, upper=179.9)
        phase.add_boundary_constraint('theta', loc='final', lower=0.01, upper=179.9)

        phase.add_boundary_constraint('x', loc='final', equals=10)
        phase.add_boundary_constraint('y', loc='final', equals=5)
        # Minimize time at the end of the phase
        phase.add_objective('time_phase', loc='final', scaler=10)

        p.set_solver_print(0)

        p.setup()

        phase.set_time_val(initial=0.0, duration=1.5)

        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])

        phase.set_control_val('theta', [5, 100])
        phase.set_parameter_val('g', 9.80665)

        p.run_driver()
        assert_near_equal(p.get_val('traj.phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-4)

    def test_brachistochrone_ode_prefix(self):

        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.options['optimizer'] = 'SLSQP'
        p.driver.declare_coloring(tol=1.0E-12)

        tx = dm.Birkhoff(num_nodes=25, grid_type='lgl')

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)
        phase.timeseries_options['use_prefix'] = True
        p.model.add_subsystem('traj', traj)
        traj.add_phase('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=False)
        phase.add_state('y', fix_initial=True, fix_final=False)

        # Note that by omitting the targets here Dymos will automatically attempt to connect
        # to a top-level input named 'v' in the ODE, and connect to nothing if it's not found.
        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta',
                          continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', targets=['g'], units='m/s**2')

        phase.add_boundary_constraint('x', loc='final', equals=10)
        phase.add_boundary_constraint('y', loc='final', equals=5)
        phase.add_boundary_constraint('check', loc='final', lower=-50, upper=50)
        phase.add_path_constraint('check', upper=100, lower=-100)
        # Minimize time at the end of the phase
        phase.add_objective('time_phase', loc='final', scaler=10)

        p.set_solver_print(0)

        p.setup()

        phase.set_time_val(initial=0.0, duration=1.5)

        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])

        phase.set_control_val('theta', [5, 100])
        phase.set_parameter_val('g', 9.80665)

        p.run_driver()
        assert_near_equal(p.get_val('traj.phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-4)

    def test_brachistochrone_ode_no_prefix(self):

        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.options['optimizer'] = 'SLSQP'
        p.driver.declare_coloring(tol=1.0E-12)

        tx = dm.Birkhoff(num_nodes=25, grid_type='lgl')

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)
        phase.timeseries_options['use_prefix'] = False
        p.model.add_subsystem('traj', traj)
        traj.add_phase('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=False)
        phase.add_state('y', fix_initial=True, fix_final=False)

        # Note that by omitting the targets here Dymos will automatically attempt to connect
        # to a top-level input named 'v' in the ODE, and connect to nothing if it's not found.
        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta',
                          continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', targets=['g'], units='m/s**2')

        phase.add_boundary_constraint('x', loc='final', equals=10)
        phase.add_boundary_constraint('y', loc='final', equals=5)
        phase.add_boundary_constraint('check', loc='final', lower=-50, upper=50)
        phase.add_path_constraint('check', upper=100, lower=-100)

        # Minimize time at the end of the phase
        phase.add_objective('time_phase', loc='final', scaler=10)

        p.set_solver_print(0)

        p.setup()

        phase.set_time_val(initial=0.0, duration=1.5)

        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])

        phase.set_control_val('theta', [5, 100])
        phase.set_parameter_val('g', 9.80665)

        p.run_driver()
        assert_near_equal(p.get_val('traj.phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-4)



import openmdao.api as om
import numpy as np


class NormalForceODE(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # constants
        self.add_input('M', val=0.0, desc='mass', units='kg')
        self.add_input('g', val=9.8, desc='mass', units='m/s**2')
        self.add_input('a', val=1.8, desc='cg to front distance', units='m')
        self.add_input('b', val=1.6, desc='cg to rear distance', units='m')
        self.add_input('tw', val=0.73, desc='half track width', units='m')
        self.add_input('h', val=0.0, desc='cg height', units='m')  # changed from 0.4
        self.add_input('chi', val=0.0, desc='roll stiffness', units=None)
        self.add_input('CoP', val=1.8, desc='center of pressure (distance to front)', units='m')

        # adding downforce
        self.add_input('ClA', val=4.0, desc='lift coefficient', units='m**2')
        self.add_input('rho', val=1.2, desc='air density', units='kg/m**3')
        self.add_input('V', val=np.zeros(nn), desc='speed', units='m/s')

        # states
        self.add_input('ax', val=np.zeros(nn), desc='longitudinal acceleration', units='m/s**2')
        self.add_input('ay', val=np.zeros(nn), desc='lateral acceleration', units='m/s**2')

        # normal load outputs
        self.add_output('N_fl', val=np.zeros(nn), desc='normal force fl', units='N')
        self.add_output('N_fr', val=np.zeros(nn), desc='normal force fr', units='N')
        self.add_output('N_rl', val=np.zeros(nn), desc='normal force rl', units='N')
        self.add_output('N_rr', val=np.zeros(nn), desc='normal force rr', units='N')

        # Setup partials
        arange = np.arange(self.options['num_nodes'], dtype=int)

        # partials
        self.declare_partials(of='N_fl', wrt='ax', rows=arange, cols=arange)
        self.declare_partials(of='N_fr', wrt='ax', rows=arange, cols=arange)
        self.declare_partials(of='N_rl', wrt='ax', rows=arange, cols=arange)
        self.declare_partials(of='N_rr', wrt='ax', rows=arange, cols=arange)

        self.declare_partials(of='N_fl', wrt='ay', rows=arange, cols=arange)
        self.declare_partials(of='N_fr', wrt='ay', rows=arange, cols=arange)
        self.declare_partials(of='N_rl', wrt='ay', rows=arange, cols=arange)
        self.declare_partials(of='N_rr', wrt='ay', rows=arange, cols=arange)

        self.declare_partials(of='N_fl', wrt='V', rows=arange, cols=arange)
        self.declare_partials(of='N_fr', wrt='V', rows=arange, cols=arange)
        self.declare_partials(of='N_rl', wrt='V', rows=arange, cols=arange)
        self.declare_partials(of='N_rr', wrt='V', rows=arange, cols=arange)

    def compute(self, inputs, outputs):
        M = inputs['M']
        g = inputs['g']
        a = inputs['a']
        b = inputs['b']
        ax = inputs['ax']
        ay = inputs['ay']
        h = inputs['h']
        chi = inputs['chi']
        tw = inputs['tw']
        rho = inputs['rho']
        V = inputs['V']
        CoP = inputs['CoP']
        ClA = inputs['ClA']

        downforce = 0.5 * rho * ClA * V ** 2

        downforce_rear = downforce * (CoP / (a + b))
        downforce_front = downforce * (1 - (CoP / (a + b)))

        outputs['N_fl'] = (M * g / 2) * (b / (a + b)) + (M / 4) * (
                          (-(ax * h) / (a + b)) + (ay * chi * h / tw)) + downforce_front / 2
        outputs['N_fr'] = (M * g / 2) * (b / (a + b)) + (M / 4) * (
                          (-(ax * h) / (a + b)) - (ay * chi * h / tw)) + downforce_front / 2
        outputs['N_rl'] = (M * g / 2) * (a / (a + b)) + (M / 4) * (
                          ((ax * h) / (a + b)) + (ay * (1 - chi) * h / tw)) + downforce_rear / 2
        outputs['N_rr'] = (M * g / 2) * (a / (a + b)) + (M / 4) * (
                          ((ax * h) / (a + b)) - (ay * (1 - chi) * h / tw)) + downforce_rear / 2

    def compute_partials(self, inputs, jacobian):
        M = inputs['M']
        a = inputs['a']
        b = inputs['b']
        h = inputs['h']
        chi = inputs['chi']
        tw = inputs['tw']
        rho = inputs['rho']
        V = inputs['V']
        CoP = inputs['CoP']
        ClA = inputs['ClA']

        jacobian['N_fl', 'ax'] = -(M * h) / (4 * (a + b))
        jacobian['N_fr', 'ax'] = -(M * h) / (4 * (a + b))
        jacobian['N_rl', 'ax'] = (M * h) / (4 * (a + b))
        jacobian['N_rr', 'ax'] = (M * h) / (4 * (a + b))

        jacobian['N_fl', 'ay'] = (M * chi * h) / (4 * tw)
        jacobian['N_rl', 'ay'] = (M * (1 - chi) * h) / (4 * tw)
        jacobian['N_fr', 'ay'] = -(M * chi * h) / (4 * tw)
        jacobian['N_rr', 'ay'] = -(M * (1 - chi) * h) / (4 * tw)

        ddownforce_dv = rho * ClA * V

        jacobian['N_fl', 'V'] = ddownforce_dv * (1 - (CoP / (a + b))) * 0.5
        jacobian['N_fr', 'V'] = ddownforce_dv * (1 - (CoP / (a + b))) * 0.5
        jacobian['N_rl', 'V'] = ddownforce_dv * (CoP / (a + b)) * 0.5
        jacobian['N_rr', 'V'] = ddownforce_dv * (CoP / (a + b)) * 0.5



import os
import unittest

try:
    import matplotlib.pyplot as plt
    plt.switch_backend('Agg')
    plt.style.use('ggplot')
except ImportError:
    plt = None

import numpy as np

from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse


@use_tempdirs
class TestReentryForDocs(unittest.TestCase):

    def tearDown(self):
        for filename in ['total_coloring.pkl', 'SLSQP.out', 'SNOPT_print.out', 'SNOPT_summary.out']:
            if os.path.exists(filename):
                os.remove(filename)

    @require_pyoptsparse(optimizer='SLSQP')
    @unittest.skipIf(plt is None, "This test requires matplotlib")
    def test_reentry(self):
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm
        from dymos.examples.shuttle_reentry.shuttle_ode import ShuttleODE
        from dymos.examples.plotting import plot_results

        # Instantiate the problem, add the driver, and allow it to use coloring
        p = om.Problem(model=om.Group())
        p.driver = om.pyOptSparseDriver()
        p.driver.declare_coloring()
        p.driver.options['optimizer'] = 'SLSQP'

        # Instantiate the trajectory and add a phase to it
        traj = p.model.add_subsystem('traj', dm.Trajectory())
        phase0 = traj.add_phase('phase0',
                                dm.Phase(ode_class=ShuttleODE,
                                         transcription=dm.Radau(num_segments=15, order=3)))

        phase0.set_time_options(fix_initial=True, units='s', duration_ref=200)
        phase0.add_state('h', fix_initial=True, fix_final=True, units='ft', rate_source='hdot',
                         lower=0, ref0=75000, ref=300000, defect_ref=1000)
        phase0.add_state('gamma', fix_initial=True, fix_final=True, units='rad',
                         rate_source='gammadot',
                         lower=-89. * np.pi / 180, upper=89. * np.pi / 180)
        phase0.add_state('phi', fix_initial=True, fix_final=False, units='rad',
                         rate_source='phidot', lower=0, upper=89. * np.pi / 180)
        phase0.add_state('psi', fix_initial=True, fix_final=False, units='rad',
                         rate_source='psidot', lower=0, upper=90. * np.pi / 180)
        phase0.add_state('theta', fix_initial=True, fix_final=False, units='rad',
                         rate_source='thetadot',
                         lower=-89. * np.pi / 180, upper=89. * np.pi / 180)
        phase0.add_state('v', fix_initial=True, fix_final=True, units='ft/s',
                         rate_source='vdot', lower=0, ref0=2500, ref=25000)
        phase0.add_control('alpha', units='rad', opt=True, lower=-np.pi / 2, upper=np.pi / 2, )
        phase0.add_control('beta', units='rad', opt=True, lower=-89 * np.pi / 180, upper=1 * np.pi / 180, )

        # The original implementation by Betts includes a heating rate path constraint.
        # This will work with the SNOPT optimizer but SLSQP has difficulty converging the solution.
        # phase0.add_path_constraint('q', lower=0, upper=70, ref=70)
        phase0.add_timeseries_output('q', shape=(1,))

        phase0.add_objective('theta', loc='final', ref=-0.01)

        p.setup(check=True)

        phase0.set_time_val(initial=0, duration=2000, units='s')
        phase0.set_state_val('h', [260000, 80000], units='ft')
        phase0.set_state_val('gamma', [-1, -5], units='deg')
        phase0.set_state_val('phi', [0, 75], units='deg')
        phase0.set_state_val('psi', [90, 10], units='deg')
        phase0.set_state_val('theta', [0, 25], units='deg')
        phase0.set_state_val('v', [25600, 2500], units='ft/s')
        phase0.set_control_val('alpha', 17.4, units='deg')
        phase0.set_control_val('beta', [-75, 0], units='deg')

        # Run the driver
        dm.run_problem(p)

        # Check the validity of the solution
        assert_near_equal(p.get_val('traj.phase0.timeseries.time')[-1], 2008.59,
                          tolerance=1e-3)
        assert_near_equal(p.get_val('traj.phase0.timeseries.theta', units='deg')[-1],
                          34.1412, tolerance=1e-3)

        # Run the simulation to check if the model is physically valid
        sim_out = traj.simulate()

        # Plot the results

        plot_results([('traj.phase0.timeseries.time', 'traj.phase0.timeseries.alpha',
                       'time (s)', 'alpha (rad)'),
                      ('traj.phase0.timeseries.time', 'traj.phase0.timeseries.beta',
                       'time (s)', 'beta (rad)'),
                      ('traj.phase0.timeseries.time', 'traj.phase0.timeseries.theta',
                       'time (s)', 'theta (rad)'),
                      ('traj.phase0.timeseries.time', 'traj.phase0.timeseries.q',
                       'time (s)', 'q (btu/ft/ft/s')], title='Reentry Solution', p_sol=p,
                     p_sim=sim_out)

        plt.show()


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest
import numpy as np
import openmdao.api as om
import dymos as dm

from dymos.examples.finite_burn_orbit_raise.finite_burn_eom import FiniteBurnODE

from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse
from openmdao.utils.assert_utils import assert_near_equal


@require_pyoptsparse(optimizer='SLSQP')
@use_tempdirs
class TestTwoPhaseOrbitDoubleLinked(unittest.TestCase):

    def test_two_phase_orbit_double_linked(self):
        traj = dm.Trajectory()
        p = om.Problem(model=om.Group())
        p.model.add_subsystem('traj', traj)

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = 'SLSQP'

        p.driver.declare_coloring()

        traj.add_parameter('c', opt=False, val=0.0, units='DU/TU',
                           targets={'phase0': ['c']})

        traj.add_parameter('accel', opt=False, val=0.0, units='DU/TU**2',
                           targets={'phase0': ['accel']})

        traj.add_parameter('u1', opt=False, val=0.0, units='deg',
                           targets={'phase0': ['u1']})

        # First half of the orbit
        phase0 = dm.Phase(ode_class=FiniteBurnODE,
                          transcription=dm.GaussLobatto(num_segments=10, order=3, compressed=True))

        phase0 = traj.add_phase('phase0', phase0)

        phase0.set_time_options(fix_initial=True, fix_duration=True, units='TU')
        phase0.add_state('r', fix_initial=False, fix_final=False,
                         rate_source='r_dot', targets=['r'], units='DU')
        phase0.add_state('theta', fix_initial=True, fix_final=False,
                         rate_source='theta_dot', targets=['theta'], units='rad')
        phase0.add_state('vr', fix_initial=False, fix_final=False,
                         rate_source='vr_dot', targets=['vr'], units='DU/TU')
        phase0.add_state('vt', fix_initial=False, fix_final=False,
                         rate_source='vt_dot', targets=['vt'], units='DU/TU')

        phase0.add_objective('time', loc='final')

        # Second half of the orbit
        phase1 = dm.Phase(ode_class=FiniteBurnODE,
                          transcription=dm.GaussLobatto(num_segments=10, order=3, compressed=True))

        phase1 = traj.add_phase('phase1', phase1)

        phase1.set_time_options(fix_initial=True, fix_duration=True, units='TU')
        phase1.add_state('r', fix_initial=False, fix_final=False,
                         rate_source='r_dot', targets=['r'], units='DU')
        phase1.add_state('theta', fix_initial=False, fix_final=False,
                         rate_source='theta_dot', targets=['theta'], units='rad')
        phase1.add_state('vr', fix_initial=False, fix_final=False,
                         rate_source='vr_dot', targets=['vr'], units='DU/TU')
        phase1.add_state('vt', fix_initial=False, fix_final=False,
                         rate_source='vt_dot', targets=['vt'], units='DU/TU')

        traj.link_phases(phases=['phase0', 'phase1'], vars=['r', 'theta', 'vr', 'vt'], connected=False)
        traj.link_phases(phases=['phase1', 'phase0'], vars=['r', 'vr', 'vt'], connected=False)

        p.setup()
        phase0.set_time_val(initial=0.0, duration=np.pi)
        phase0.set_state_val('r', vals=[2, 1])
        phase0.set_state_val('theta', vals=[0, np.pi])
        phase0.set_state_val('vr', vals=[0, 0])
        phase0.set_state_val('vt', vals=[2, 1])

        phase1.set_time_val(initial=np.pi, duration=np.pi)
        phase0.set_state_val('r', vals=[1, 1])
        phase0.set_state_val('theta', vals=[0, np.pi])
        phase0.set_state_val('vr', vals=[0, 0])
        phase0.set_state_val('vt', vals=[1, 1])

        dm.run_problem(p, make_plots=True)

        r_1 = p.get_val('traj.phase0.timeseries.r')
        theta_2 = p.get_val('traj.phase1.timeseries.theta')

        assert_near_equal(r_1[0], 1.0, tolerance=1e-9)
        assert_near_equal(theta_2[-1], 2*np.pi, tolerance=1e-9)



"""
Utility for storing and outputting grid refinement information at each iteration
"""
import numpy as np


def write_error(f, iter_number, phases, refine_results):
    """
    Writes a summary of the current grid refinement iteration to the given stream.

    Parameters
    ----------
    f : stream
        The output stream to which the grid refinment should be printed.
    iter_number : int
        The current grid refinement iteration index.
    phases : dict of {phase_path: Phase}
        The phases in the problem being refined.
    refine_results : dict
        A dictionary containing the grid refinement data for each phase, keyed by the phase path
        in the model.
    """
    f.write('\n\n')
    print(50 * '=', file=f)
    str_gr = f'Grid Refinement - Iteration {iter_number}'
    print(f'{str_gr:^50}', file=f)
    print(50 * '-', file=f)
    for phase_path, phase in phases.items():
        refine_data = refine_results[phase_path]
        refine_options = phase.refine_options

        f.write('    Phase: {}\n'.format(phase_path))

        # Print the phase grid-refinement settings
        print('        Refinement Options:', file=f)
        print('            Allow Refinement = {}'.format(refine_options['refine']), file=f)
        print('            Tolerance = {}'.format(refine_options['tolerance']), file=f)
        print('            Min Order = {}'.format(refine_options['min_order']), file=f)
        print('            Max Order = {}'.format(refine_options['max_order']), file=f)

        # Print the original grid specs
        print('        Original Grid:', file=f)
        print('            Number of Segments = {}'.format(refine_data['num_segments']), file=f)

        str_segends = ', '.join(str(round(elem, 4)) for elem in refine_data['segment_ends'])
        print(f'            Segment Ends = [{str_segends}]', file=f)

        str_segorders = ', '.join(str(elem) for elem in refine_data['order'])
        print(f'            Segment Order = [{str_segorders}]', file=f)

        error = refine_data['max_rel_error']
        str_errors = ', '.join(f'{elem:8.4g}' for elem in error)
        print(f'            Error = [{str_errors}]', file=f)

    return


def write_refine_iter(f, iter_number, phases, refine_results):
    """
    Writes a summary of the current grid refinement iteration to the given stream.

    Parameters
    ----------
    f : stream
        The output stream to which the grid refinment should be printed.
    iter_number : int
        The current grid refinement iteration index.
    phases : dict of {phase_path: Phase}
        The phases in the problem being refined.
    refine_results : dict
        A dictionary containing the grid refinement data for each phase, keyed by the phase path
        in the model.
    """

    for phase_path, phase in phases.items():
        refine_data = refine_results[phase_path]
        refine_options = phase.refine_options

        f.write('    Phase: {}\n'.format(phase_path))
        # Print the modified grid specs
        print('        New Grid:', file=f)
        print('            Number of Segments = {}'.format(refine_data['new_num_segments']), file=f)

        str_segends = ', '.join(str(round(elem, 4)) for elem in refine_data['new_segment_ends'])
        print(f'            Segment Ends = [{str_segends}]', file=f)

        new_order = refine_data['new_order']
        str_segorders = ', '.join(str(elem) for elem in new_order)

        print(f'            Segment Order = [{str_segorders}]', file=f)

        is_refined = True if np.any(refine_data['need_refinement']) and refine_options['refine'] else False
        print(f'        Refined: {is_refined}', file=f)
        print(file=f)
    return



import openmdao.api as om
import numpy as np


class AccelerationODE(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # constants
        self.add_input('tau_y', val=0.2, desc='lateral load transfer time constant', units='s')
        self.add_input('tau_x', val=0.2, desc='longitudinal load transfer time constant', units='s')

        # states
        self.add_input('V', val=np.zeros(nn), desc='speed', units='m/s')
        self.add_input('lambda', val=np.zeros(nn), desc='body slip angle', units='rad')
        self.add_input('omega', val=np.zeros(nn), desc='yaw rate', units='rad/s')

        self.add_input('Vdot', val=np.zeros(nn), desc='speed', units='m/s**2')
        self.add_input('lambdadot', val=np.zeros(nn), desc='body slip angle', units='rad/s')

        self.add_input('ax', val=np.zeros(nn), desc='longitudinal acceleration', units='m/s**2')
        self.add_input('ay', val=np.zeros(nn), desc='lateral acceleration', units='m/s**2')

        # outputs
        self.add_output('axdot', val=np.zeros(nn), desc='longitudinal jerk', units='m/s**3')
        self.add_output('aydot', val=np.zeros(nn), desc='lateral jerk', units='m/s**3')

        # Setup partials
        arange = np.arange(self.options['num_nodes'], dtype=int)

        # partials
        self.declare_partials(of='axdot', wrt='ax', rows=arange, cols=arange)
        self.declare_partials(of='axdot', wrt='Vdot', rows=arange, cols=arange)
        self.declare_partials(of='axdot', wrt='omega', rows=arange, cols=arange)
        self.declare_partials(of='axdot', wrt='V', rows=arange, cols=arange)
        self.declare_partials(of='axdot', wrt='lambda', rows=arange, cols=arange)

        self.declare_partials(of='aydot', wrt='ay', rows=arange, cols=arange)
        self.declare_partials(of='aydot', wrt='omega', rows=arange, cols=arange)
        self.declare_partials(of='aydot', wrt='V', rows=arange, cols=arange)
        self.declare_partials(of='aydot', wrt='Vdot', rows=arange, cols=arange)
        self.declare_partials(of='aydot', wrt='lambda', rows=arange, cols=arange)
        self.declare_partials(of='aydot', wrt='lambdadot', rows=arange, cols=arange)

    def compute(self, inputs, outputs):
        tau_y = inputs['tau_y']
        tau_x = inputs['tau_x']
        V = inputs['V']
        lamb = inputs['lambda']
        omega = inputs['omega']
        Vdot = inputs['Vdot']
        lambdadot = inputs['lambdadot']
        ax = inputs['ax']
        ay = inputs['ay']

        outputs['axdot'] = (Vdot+omega*V*lamb-ax)/tau_x
        outputs['aydot'] = (omega*V-(V*lambdadot+Vdot*lamb)-ay)/tau_y

    def compute_partials(self, inputs, jacobian):
        tau_y = inputs['tau_y']
        tau_x = inputs['tau_x']
        V = inputs['V']
        lamb = inputs['lambda']
        omega = inputs['omega']
        Vdot = inputs['Vdot']
        lambdadot = inputs['lambdadot']

        jacobian['axdot', 'ax'] = -1/tau_x
        jacobian['axdot', 'Vdot'] = 1/tau_x
        jacobian['axdot', 'omega'] = (V*lamb)/tau_x
        jacobian['axdot', 'lambda'] = (omega*V)/tau_x
        jacobian['axdot', 'V'] = (omega*lamb)/tau_x

        jacobian['aydot', 'ay'] = -1/tau_y
        jacobian['aydot', 'omega'] = V/tau_y
        jacobian['aydot', 'V'] = (omega-lambdadot)/tau_y
        jacobian['aydot', 'lambda'] = -Vdot/tau_y
        jacobian['aydot', 'lambdadot'] = -V/tau_y
        jacobian['aydot', 'Vdot'] = -lamb/tau_y



import unittest

import numpy as np
from numpy.testing import assert_almost_equal

from dymos.utils.lgr import lgr

# Known solutions
x_i = {2: [-1.0, 1 / 3.0],
       3: [-1.0, -0.289898, 0.689898],
       4: [-1.0, -0.575319, 0.181066, 0.822824],
       5: [-1.0, -0.72048, -0.167181, 0.446314, 0.885792]}

w_i = {2: [0.5, 1.5],
       3: [2.0 / 9.0, (16 + np.sqrt(6)) / 18.0, (16 - np.sqrt(6)) / 18.0],
       4: [0.125, 0.657689, 0.776387, 0.440924],
       5: [0.08, 0.446208, 0.623653, 0.562712, 0.287427]}


class TestLGR(unittest.TestCase):

    def test_nodes_and_weights_2(self):
        x_2, w_2 = lgr(2)
        assert_almost_equal(x_2, x_i[2], decimal=6)
        assert_almost_equal(w_2, w_i[2], decimal=6)

        x_2, w_2 = lgr(2, include_endpoint=True)
        assert_almost_equal(x_2, x_i[2] + [1], decimal=6)
        assert_almost_equal(w_2, w_i[2] + [0], decimal=6)

    def test_nodes_and_weights_3(self):
        x_3, w_3 = lgr(3)
        assert_almost_equal(x_3, x_i[3], decimal=6)
        assert_almost_equal(w_3, w_i[3], decimal=6)

        x_3, w_3 = lgr(3, include_endpoint=True)
        assert_almost_equal(x_3, x_i[3] + [1], decimal=6)
        assert_almost_equal(w_3, w_i[3] + [0], decimal=6)

    def test_nodes_and_weights_4(self):
        x_4, w_4 = lgr(4)
        assert_almost_equal(x_4, x_i[4], decimal=6)
        assert_almost_equal(w_4, w_i[4], decimal=6)

        x_4, w_4 = lgr(4, include_endpoint=True)
        assert_almost_equal(x_4, x_i[4] + [1], decimal=6)
        assert_almost_equal(w_4, w_i[4] + [0], decimal=6)

    def test_nodes_and_weights_5(self):
        x_5, w_5 = lgr(5)
        assert_almost_equal(x_5, x_i[5], decimal=6)
        assert_almost_equal(w_5, w_i[5], decimal=6)

        x_5, w_5 = lgr(5, include_endpoint=True)
        assert_almost_equal(x_5, x_i[5] + [1], decimal=6)
        assert_almost_equal(w_5, w_i[5] + [0], decimal=6)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import numpy as np
import scipy.sparse as sp
import openmdao.api as om
from ...grid_data import GridData
from ....utils.misc import get_rate_units
from ...._options import options as dymos_options


class StateInterpComp(om.ExplicitComponent):
    r"""
    Provide interpolated state values and/or rates for pseudospectral transcriptions.

    When the transcription is *gauss-lobatto* it accepts the state values and derivatives
    at discretization nodes and computes the interpolated state values and derivatives
    at the collocation nodes, using a Hermite interpolation scheme.

    .. math:: x_c = \left[ A_i \right] x_d + \frac{dt}{d\tau_s} \left[ B_i \right] f_d
    .. math:: \dot{x}_c = \frac{d\tau_s}{dt} \left[ A_d \right] x_d + \left[ B_d \right] f_d

    When the transcription is *radau-ps* it accepts the state values at the discretization nodes
    and computes the interpolated state derivatives at the collocation nodes, using a Lagrange
    interpolation scheme.

    .. math:: \dot{x}_c = \frac{d\tau_s}{dt} \left[ A_d \right] x_d

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._no_check_partials = not dymos_options['include_check_partials']

    def initialize(self):
        """
        Declare component options.
        """
        self.options.declare(
            'transcription', values=['gauss-lobatto', 'radau-ps'],
            desc='Transcription technique of the optimal control problem.')

        self.options.declare(
            'grid_data', types=GridData,
            desc='Container object for grid info')

        self.options.declare(
            'state_options', types=dict,
            desc='Dictionary of state names/options for the phase')

        self.options.declare(
            'time_units', default=None, allow_none=True, types=str,
            desc='Units of the integration variable')

    def configure_io(self):
        """
        I/O creation is delayed until configure so we can determine shape and units.
        """
        time_units = self.options['time_units']

        num_disc_nodes = self.options['grid_data'].subset_num_nodes['state_disc']
        num_col_nodes = self.options['grid_data'].subset_num_nodes['col']

        state_options = self.options['state_options']

        transcription = self.options['transcription']

        self.add_input(name='dt_dstau', shape=(num_col_nodes,), units=time_units,
                       desc='For each node, the duration of its '
                            'segment in the integration variable')

        self.xd_str = {}
        self.fd_str = {}
        self.xc_str = {}
        self.xdotc_str = {}

        for state_name, options in state_options.items():
            shape = options['shape']
            units = options['units']

            rate_units = get_rate_units(units, time_units)

            self.add_input(
                name=f'state_disc:{state_name}',
                shape=(num_disc_nodes,) + shape,
                desc=f'Values of state {state_name} at discretization nodes',
                units=units)

            if transcription == 'gauss-lobatto':
                self.add_input(
                    name=f'staterate_disc:{state_name}',
                    shape=(num_disc_nodes,) + shape,
                    units=rate_units,
                    desc='EOM time derivative of state {state_name} at discretization nodes')

                self.add_output(
                    name=f'state_col:{state_name}',
                    shape=(num_col_nodes,) + shape, units=units,
                    desc=f'Interpolated values of state {state_name} at collocation nodes')

            self.add_output(
                name=f'staterate_col:{state_name}',
                shape=(num_col_nodes,) + shape,
                units=rate_units,
                desc=f'Interpolated rate of state {state_name} at collocation nodes')

            self.xd_str[state_name] = f'state_disc:{state_name}'
            self.fd_str[state_name] = f'staterate_disc:{state_name}'
            self.xc_str[state_name] = f'state_col:{state_name}'
            self.xdotc_str[state_name] = f'staterate_col:{state_name}'

        if transcription == 'gauss-lobatto':
            Ai, Bi, Ad, Bd = self.options['grid_data'].phase_hermite_matrices('state_disc', 'col', sparse=True)

        elif transcription == 'radau-ps':
            Ai, Ad = self.options['grid_data'].phase_lagrange_matrices('state_disc', 'col', sparse=True)
            Bi = Bd = sp.csr_matrix(np.zeros(shape=(num_col_nodes, num_disc_nodes)))

        else:
            raise ValueError(f"unhandled transcription type: {self.options['transcription']}")

        self.matrices = {'Ai': Ai, 'Bi': Bi, 'Ad': Ad, 'Bd': Bd}
        self.jacs = {'Ai': {}, 'Bi': {}, 'Ad': {}, 'Bd': {}}
        self.sizes = {}

        for name, options in state_options.items():
            shape = options['shape']

            size = np.prod(shape)
            self.sizes[name] = size

            for key in self.jacs:
                # Each jacobian matrix has a form that is defined by the Kronecker product
                # of the interpolation matrix eye(size).
                self.jacs[key][name] = sp.kron(sp.csr_matrix(self.matrices[key]),
                                               sp.eye(size, format='csr'),
                                               format='csr')

            self.sizes[name] = size

            #
            # Partial of xdotc wrt dt_dstau
            #
            rs = np.arange(num_col_nodes * size, dtype=int)
            cs = np.repeat(np.arange(num_col_nodes, dtype=int), size)

            self.declare_partials(of=self.xdotc_str[name], wrt='dt_dstau',
                                  rows=rs, cols=cs)

            if transcription == 'gauss-lobatto':
                self.declare_partials(
                    of=self.xc_str[name], wrt='dt_dstau',
                    rows=rs, cols=cs)

                Ai_rows, Ai_cols, data = sp.find(self.jacs['Ai'][name])
                self.declare_partials(of=self.xc_str[name], wrt=self.xd_str[name],
                                      rows=Ai_rows, cols=Ai_cols, val=data)

                Bi_rows, Bi_cols = self.jacs['Bi'][name].nonzero()
                self.declare_partials(of=self.xc_str[name], wrt=self.fd_str[name],
                                      rows=Bi_rows, cols=Bi_cols)

                Bd_rows, Bd_cols, data = sp.find(self.jacs['Bd'][name])
                self.declare_partials(of=self.xdotc_str[name], wrt=self.fd_str[name],
                                      rows=Bd_rows, cols=Bd_cols, val=data)

            Ad_rows, Ad_cols = self.jacs['Ad'][name].nonzero()
            self.declare_partials(of=self.xdotc_str[name], wrt=self.xd_str[name],
                                  rows=Ad_rows, cols=Ad_cols)

    def _compute_radau(self, inputs, outputs):
        num_disc_nodes = self.options['grid_data'].subset_num_nodes['state_disc']
        num_col_nodes = self.options['grid_data'].subset_num_nodes['col']
        state_options = self.options['state_options']
        dt_dstau = inputs['dt_dstau'][:, np.newaxis]

        for name in state_options:
            xdotc_str = self.xdotc_str[name]
            xd_str = self.xd_str[name]
            shape = state_options[name]['shape']
            size = np.prod(shape)

            xd_flat = np.reshape(inputs[xd_str],
                                 (num_disc_nodes, size))

            outputs[xdotc_str] = np.reshape(self.matrices['Ad'].dot(xd_flat) / dt_dstau,
                                            (num_col_nodes,) + shape)

    def _compute_gauss_lobatto(self, inputs, outputs):
        state_options = self.options['state_options']

        dt_dstau = inputs['dt_dstau'][:, np.newaxis]

        num_disc_nodes = self.options['grid_data'].subset_num_nodes['state_disc']
        num_col_nodes = self.options['grid_data'].subset_num_nodes['col']

        Ai = self.matrices['Ai']
        Bi = self.matrices['Bi']
        Ad = self.matrices['Ad']
        Bd = self.matrices['Bd']

        for name in state_options:
            shape = state_options[name]['shape']
            size = np.prod(shape)
            xc_str = self.xc_str[name]
            xdotc_str = self.xdotc_str[name]
            xd_str = self.xd_str[name]
            fd_str = self.fd_str[name]

            xd_flat = np.reshape(inputs[xd_str], (num_disc_nodes, size))
            fd_flat = np.reshape(inputs[fd_str], (num_disc_nodes, size))

            col_val = Bi.dot(fd_flat) * dt_dstau + Ai.dot(xd_flat)

            outputs[xc_str] = np.reshape(col_val, (num_col_nodes,) + shape)

            col_rate = Ad.dot(xd_flat) / dt_dstau + Bd.dot(fd_flat)

            outputs[xdotc_str] = np.reshape(col_rate, (num_col_nodes,) + shape)

    def _compute_partials_radau(self, inputs, partials):
        state_options = self.options['state_options']

        ndn = self.options['grid_data'].subset_num_nodes['state_disc']

        Ad = self.matrices['Ad']

        dstau_dt = np.reciprocal(inputs['dt_dstau'])
        dstau_dt2 = (dstau_dt ** 2)

        for name in state_options:
            size = self.sizes[name]

            xdotc_name = self.xdotc_str[name]
            xd_name = self.xd_str[name]

            # Unroll matrix-shaped states into an array at each node
            xd_flat = np.reshape(inputs[xd_name], (ndn, size))

            partials[xdotc_name, 'dt_dstau'] = (-Ad.dot(xd_flat) * dstau_dt2[:, np.newaxis]).ravel()

            dstau_dt_x_size = np.repeat(dstau_dt, size)[:, np.newaxis]

            dxdotc_dxd = self.jacs['Ad'][name].multiply(dstau_dt_x_size)
            partials[xdotc_name, xd_name] = dxdotc_dxd.data

    def _compute_partials_gauss_lobatto(self, inputs, partials):
        ndn = self.options['grid_data'].subset_num_nodes['state_disc']

        Ad = self.matrices['Ad']
        Bi = self.matrices['Bi']

        dstau_dt = np.reciprocal(inputs['dt_dstau'])
        dstau_dt2 = dstau_dt ** 2

        for name in self.options['state_options']:
            size = self.sizes[name]

            xdotc_name = self.xdotc_str[name]
            xd_name = self.xd_str[name]

            xc_name = self.xc_str[name]
            fd_name = self.fd_str[name]

            # Unroll matrix-shaped states into an array at each node
            xd = np.reshape(inputs[xd_name], (ndn, size))
            fd = np.reshape(inputs[fd_name], (ndn, size))

            dt_dstau_x_size = np.repeat(inputs['dt_dstau'], size)[:, np.newaxis]

            partials[xc_name, 'dt_dstau'] = Bi.dot(fd).ravel()

            partials[xdotc_name, 'dt_dstau'] = (-Ad.dot(xd) * dstau_dt2[:, np.newaxis]).ravel()

            dxc_dfd = self.jacs['Bi'][name].multiply(dt_dstau_x_size)
            partials[xc_name, fd_name] = dxc_dfd.data

            dxdotc_dxd = self.jacs['Ad'][name].multiply(np.reciprocal(dt_dstau_x_size))
            partials[xdotc_name, xd_name] = dxdotc_dxd.data

    def compute(self, inputs, outputs):
        """
        Interpolate state values.

        Parameters
        ----------
        inputs : `Vector`
            `Vector` containing inputs.
        outputs : `Vector`
            `Vector` containing outputs.
        """
        transcription = self.options['transcription']

        if transcription == 'gauss-lobatto':
            self._compute_gauss_lobatto(inputs, outputs)
        elif transcription == 'radau-ps':
            self._compute_radau(inputs, outputs)
        else:
            raise ValueError(f'Invalid transcription: {transcription}')

    def compute_partials(self, inputs, partials):
        """
        Compute sub-jacobian parts. The model is assumed to be in an unscaled state.

        Parameters
        ----------
        inputs : Vector
            Unscaled, dimensional input variables read via inputs[key].
        partials : Jacobian
            Subjac components written to partials[output_name, input_name].
        """
        transcription = self.options['transcription']
        if transcription == 'gauss-lobatto':
            self._compute_partials_gauss_lobatto(inputs, partials)
        elif transcription == 'radau-ps':
            self._compute_partials_radau(inputs, partials)
        else:
            raise ValueError(f'Invalid transcription: {transcription}')



import unittest

import numpy as np

import openmdao.api as om
from openmdao.utils.testing_utils import use_tempdirs

import dymos as dm


class BrachistochroneVectorStatesODE(om.ExplicitComponent):
    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_input('v', val=np.zeros(nn), desc='velocity', units='m/s')
        self.add_input('g', val=9.80665 * np.ones(nn), desc='grav. acceleration', units='m/s/s')
        self.add_input('theta', val=np.zeros(nn), desc='angle of wire', units='rad')

        self.add_output('pos_dot', val=np.zeros((nn, 2)), desc='velocity components', units='m/s')
        self.add_output('vdot', val=np.zeros(nn), desc='acceleration magnitude', units='m/s**2')
        self.add_output('check', val=np.zeros(nn), desc='check solution: v/sin(theta) = constant', units='m/s')

        # Setup partials
        arange = np.arange(self.options['num_nodes'])

        self.declare_partials(of='vdot', wrt='g', rows=arange, cols=arange)
        self.declare_partials(of='vdot', wrt='theta', rows=arange, cols=arange)

        self.declare_partials('*', '*', method='cs')
        self.declare_coloring(wrt='*', method='cs', show_sparsity=False)

    def compute(self, inputs, outputs):
        theta = inputs['theta']
        cos_theta = np.cos(theta)
        sin_theta = np.sin(theta)
        g = inputs['g']
        v = inputs['v']

        outputs['vdot'] = g * cos_theta
        outputs['pos_dot'][:, 0] = v * sin_theta
        outputs['pos_dot'][:, 1] = -v * cos_theta

        outputs['check'] = v / sin_theta


@use_tempdirs
class TestAddPathConstraint(unittest.TestCase):

    def test_invalid_expression(self):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.options['optimizer'] = 'SLSQP'

        p.driver.declare_coloring()

        transcription = dm.GaussLobatto(num_segments=3,
                                        order=3,
                                        compressed=True)

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=BrachistochroneVectorStatesODE,
                         transcription=transcription)
        traj.add_phase('phase0', phase)

        p.model.add_subsystem('traj0', traj)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10), units='s')

        # can't fix final position if you're solving the segments
        phase.add_state('pos',
                        rate_source='pos_dot', units='m',
                        fix_initial=True)

        phase.add_state('v',
                        rate_source='vdot', units='m/s',
                        fix_initial=True, fix_final=False)

        phase.add_control('theta',
                          continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', val=9.80665, opt=False)

        phase.add_path_constraint(name='pos**2', lower=np.array([10, 5]))

        # test add_boundary_constraint with arrays:
        expected = "Unable to find the source 'pos**2' in the ODE."
        with self.assertRaises(ValueError) as e:
            p.setup()

        self.assertEqual(expected, str(e.exception))

    def test_duplicate_name(self):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.options['optimizer'] = 'SLSQP'

        p.driver.declare_coloring()

        transcription = dm.GaussLobatto(num_segments=3,
                                        order=3,
                                        compressed=True)

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=BrachistochroneVectorStatesODE,
                         transcription=transcription)
        traj.add_phase('phase0', phase)

        p.model.add_subsystem('traj0', traj)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10), units='s')

        # can't fix final position if you're solving the segments
        phase.add_state('pos',
                        rate_source='pos_dot', units='m',
                        fix_initial=True)

        phase.add_state('v',
                        rate_source='vdot', units='m/s',
                        fix_initial=True, fix_final=False)

        phase.add_control('theta',
                          continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', val=9.80665, opt=False)

        phase.add_path_constraint(name='pos', equals=np.array([10, 5]))

        # test add_boundary_constraint with arrays:
        with self.assertRaises(ValueError) as e:
            phase.add_path_constraint(name='pos=v**2', equals=np.array([10, 5]))

        expected = 'Cannot add new path constraint named `pos` and indices None.' \
                   ' The name `pos` is already in use as a path constraint'
        self.assertEqual(expected, str(e.exception))

    def test_duplicate_constraint(self):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.options['optimizer'] = 'SLSQP'

        p.driver.declare_coloring()

        transcription = dm.GaussLobatto(num_segments=3,
                                        order=3,
                                        compressed=True)

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=BrachistochroneVectorStatesODE,
                         transcription=transcription)
        traj.add_phase('phase0', phase)

        p.model.add_subsystem('traj0', traj)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10), units='s')

        # can't fix final position if you're solving the segments
        phase.add_state('pos',
                        rate_source='pos_dot', units='m',
                        fix_initial=True)

        phase.add_state('v',
                        rate_source='vdot', units='m/s',
                        fix_initial=True, fix_final=False)

        phase.add_control('theta',
                          continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', val=9.80665, opt=False)

        phase.add_path_constraint(name='pos', equals=np.array([10, 5]))

        # test add_boundary_constraint with arrays:
        with self.assertRaises(ValueError) as e:
            phase.add_path_constraint(name='pos', equals=np.array([10, 5]))

        expected = 'Cannot add new path constraint for variable `pos` and indices None. One already exists.'
        self.assertEqual(expected, str(e.exception))



import numpy as np

import openmdao.api as om


class ThrustComp(om.ExplicitComponent):
    """ Compute thrust from the thrust coefficient
    """
    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # Inputs
        self.add_input(name='CT',
                       shape=(nn,),
                       desc='thrust coefficient',
                       units=None)

        self.add_input(name='q',
                       shape=(nn,),
                       desc='dynamic pressure',
                       units='Pa')

        self.add_input(name='S',
                       shape=(nn,),
                       desc='reference area',
                       units='m**2')

        # Outputs
        self.add_output(name='thrust',
                        val=np.zeros(nn),
                        desc='thrust',
                        units='N')

        # Partials
        ar = np.arange(nn)
        self.declare_partials('thrust', 'CT', rows=ar, cols=ar)
        self.declare_partials('thrust', 'q', rows=ar, cols=ar)
        self.declare_partials('thrust', 'S', rows=ar, cols=ar)

    def compute(self, inputs, outputs):
        outputs['thrust'] = inputs['CT'] * inputs['q'] * inputs['S']

    def compute_partials(self, inputs, partials):
        partials['thrust', 'CT'] = inputs['q'] * inputs['S']
        partials['thrust', 'q'] = inputs['CT'] * inputs['S']
        partials['thrust', 'S'] = inputs['CT'] * inputs['q']
